"""Objects for volume groups."""

__all__ = ["VolumeGroup", "VolumeGroups"]

from typing import Iterable, Union

from . import check, ObjectField, ObjectFieldRelatedSet, ObjectSet, ObjectType
from .nodes import Node
from .block_devices import BlockDevice
from .partitions import Partition
from .filesystem_groups import FilesystemGroup, DevicesField


class VolumeGroupType(ObjectType):
    """Metaclass for `VolumeGroup`."""

    async def read(cls, node, id):
        """Get `VolumeGroup` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


class VolumeGroup(FilesystemGroup, metaclass=VolumeGroupType):
    """A volume group on a machine."""

    uuid = ObjectField.Checked("uuid", check(str), check(str))

    size = ObjectField.Checked("size", check(int), check(int), readonly=True)
    available_size = ObjectField.Checked("available_size", check(int), readonly=True)
    used_size = ObjectField.Checked("used_size", check(int), readonly=True)

    devices = DevicesField("devices")
    logical_volumes = ObjectFieldRelatedSet(
        "logical_volumes", "LogicalVolumes", reverse=None
    )

    def __repr__(self):
        return super(VolumeGroup, self).__repr__(fields={"name", "size"})

    async def delete(self):
        """Delete this volume group."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)


class VolumeGroupsType(ObjectType):
    """Metaclass for `VolumeGroups`."""

    async def read(cls, node):
        """Get list of `VolumeGroup`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls,
        node: Union[Node, str],
        name: str,
        devices: Iterable[Union[BlockDevice, Partition]],
        *,
        uuid: str = None
    ):
        """
        Create a volume group on a Node.

        :param node: Node to create the interface on.
        :type node: `Node` or `str`
        :param name: Name of the volume group.
        :type name: `str`
        :param devices: Mixed list of block devices or partitions to create
            the volume group from.
        :type devices: iterable of mixed type of `BlockDevice` or `Partition`
        :param uuid: The UUID for the volume group (optional).
        :type uuid: `str`
        """
        params = {"name": name}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )

        if len(devices) == 0:
            raise ValueError("devices must contain at least one device.")

        block_devices = []
        partitions = []
        for idx, device in enumerate(devices):
            if isinstance(device, BlockDevice):
                block_devices.append(device.id)
            elif isinstance(device, Partition):
                partitions.append(device.id)
            else:
                raise TypeError(
                    "devices[%d] must be a BlockDevice or "
                    "Partition, not %s" % type(device).__name__
                )
        if len(block_devices) > 0:
            params["block_devices"] = block_devices
        if len(partitions) > 0:
            params["partitions"] = partitions
        if uuid is not None:
            params["uuid"] = uuid
        return cls._object(await cls._handler.create(**params))


class VolumeGroups(ObjectSet, metaclass=VolumeGroupsType):
    """The set of volume groups on a machine."""

    @property
    def by_name(self):
        """Return mapping of name to `VolumeGroup`."""
        return {vg.name: vg for vg in self}

    def get_by_name(self, name):
        """Return a `VolumeGroup` by its name."""
        return self.by_name[name]



"""Commands for machines."""

__all__ = ["register"]

import asyncio
import base64
import os
import re
import subprocess
import sys
import time

from . import (
    colorized,
    CommandError,
    OriginCommand,
    OriginPagedTableCommand,
    tables,
    yes_or_no,
)
from .. import utils
from ..bones import CallError
from ..enum import NodeStatus
from ..utils.maas_async import asynchronous


def validate_file(parser, arg):
    """Validates that `arg` is a valid file."""
    if not os.path.isfile(arg):
        parser.error("%s is not a file." % arg)
    return arg


def base64_file(filepath):
    """Read from `filepath` and convert to base64."""
    with open(filepath, "rb") as stream:
        return base64.b64encode(stream.read())


class cmd_machines(OriginPagedTableCommand):
    """List machines."""

    def __init__(self, parser):
        super(cmd_machines, self).__init__(parser)
        parser.add_argument("hostname", nargs="*", help=("Hostname of the machine."))
        parser.add_argument(
            "--owned", action="store_true", help=("Show only machines owned by you.")
        )

    def execute(self, origin, options, target):
        hostnames = None
        if options.hostname:
            hostnames = options.hostname
        machines = origin.Machines.read(hostnames=hostnames)
        if options.owned:
            me = origin.Users.whoami()
            machines = origin.Machines(
                [
                    machine
                    for machine in machines
                    if machine.owner is not None
                    and machine.owner.username == me.username
                ]
            )
        table = tables.MachinesTable()
        return table.render(target, machines)


class cmd_machine(OriginPagedTableCommand):
    """Details of a machine."""

    def __init__(self, parser):
        super(cmd_machine, self).__init__(parser)
        parser.add_argument("hostname", nargs=1, help=("Hostname of the machine."))

    def execute(self, origin, options, target):
        machines = origin.Machines.read(hostnames=options.hostname)
        if len(machines) == 0:
            raise CommandError("Unable to find machine %s." % options.hostname[0])
        machine = machines[0]
        table = tables.MachineDetail()
        return table.render(target, machine)


class cmd_allocate(OriginCommand):
    """Allocate an available machine.

    Parameters can be used to allocate a machine that possesses
    certain characteristics.  All the parameters are optional and when
    multiple parameters are provided, they are combined using 'AND'
    semantics.

    Most parameters map to greater than or equal matching (e.g. --cpus 2, will
    match any available machine with 2 or more cpus). MAAS uses a cost
    algorithm to pick the machine based on the parameters that has the lowest
    usage costs to the remaining availablilty of machines.

    Parameter details:

    --disk

      The machine must have a disk present that is at least X GB in size and
      have the specified tags. The format of the parameter is:

      [<label>:]<size>[(<tag>[,<tag>])]

      Size is specified in GB. Tags map to the tags on the disk. When tags are
      included the disk must be at least X GB in size and have the specified
      tags. Label is only used in the resulting acquired machine to provide a
      mapping between the disk parameter and the disk on the machine that
      matched that parameter.

    --interface

      Machines must have one or more interfaces. The format of the parameter
      is:

      [<label>:]<key>=<value>[,<key2>=<value2>[,...]]

      Each key can be one of the following:

      - id: Matches an interface with the specific id
      - fabric: Matches an interface attached to the specified fabric.
      - fabric_class: Matches an interface attached to a fabric
        with the specified class.
      - ip: Matches an interface with the specified IP address
        assigned to it.
      - mode: Matches an interface with the specified mode. (Currently,
        the only supported mode is "unconfigured".)
      - name: Matches an interface with the specified name.
        (For example, "eth0".)
      - hostname: Matches an interface attached to the node with
        the specified hostname.
      - subnet: Matches an interface attached to the specified subnet.
      - space: Matches an interface attached to the specified space.
      - subnet_cidr: Matches an interface attached to the specified
        subnet CIDR. (For example, "192.168.0.0/24".)
      - type: Matches an interface of the specified type. (Valid
        types: "physical", "vlan", "bond", "bridge", or "unknown".)
      - vlan: Matches an interface on the specified VLAN.
      - vid: Matches an interface on a VLAN with the specified VID.
      - tag: Matches an interface tagged with the specified tag.

    --subnet

      The machine must be configured to acquire an address
      in the specified subnet, have a static IP address in the specified
      subnet, or have been observed to DHCP from the specified subnet
      during commissioning time (which implies that it *could* have an
      address on the specified subnet).

      Subnets can be specified by one of the following criteria:

      - <id>: match the subnet by its 'id' field
      - fabric:<fabric-spec>: match all subnets in a given fabric.
      - ip:<ip-address>: Match the subnet containing <ip-address> with
        the with the longest-prefix match.
      - name:<subnet-name>: Match a subnet with the given name.
      - space:<space-spec>: Match all subnets in a given space.
      - vid:<vid-integer>: Match a subnet on a VLAN with the specified
        VID. Valid values range from 0 through 4094 (inclusive). An
        untagged VLAN can be specified by using the value "0".
      - vlan:<vlan-spec>: Match all subnets on the given VLAN.

      Note that (as of this writing), the 'fabric', 'space', 'vid', and
      'vlan' specifiers are only useful for the 'not_spaces' version of
      this constraint, because they will most likely force the query
      to match ALL the subnets in each fabric, space, or VLAN, and thus
      not return any nodes. (This is not a particularly useful behavior,
      so may be changed in the future.)

      If multiple subnets are specified, the machine must be associated
      with all of them.
    """

    def __init__(
        self, parser, with_hostname=True, with_comment=True, with_dry_run=True
    ):
        super(cmd_allocate, self).__init__(parser)
        if with_hostname:
            parser.add_argument(
                "hostname", nargs="?", help=("Hostname of the machine.")
            )
        parser.add_argument(
            "--arch",
            nargs="*",
            help=(
                "Architecture(s) of the desired machine (e.g. 'i386/generic', "
                "'amd64', 'armhf/highbank', etc.)"
            ),
        )
        parser.add_argument(
            "--cpus", type=int, help=("Minimum number of CPUs for the desired machine.")
        )
        parser.add_argument(
            "--disk", nargs="*", help=("Disk(s) the desired machine must match.")
        )
        parser.add_argument(
            "--fabric",
            nargs="*",
            help=("Fabric(s) the desired machine must be connected to."),
        )
        parser.add_argument(
            "--interface",
            nargs="*",
            help=("Interface(s) the desired machine must match."),
        )
        parser.add_argument(
            "--memory",
            type=float,
            help=(
                "Minimum amount of memory (expressed in MB) for the desired " "machine."
            ),
        )
        parser.add_argument(
            "--pod", help=("Pod the desired machine must be located in.")
        )
        parser.add_argument(
            "--pod-type", help=("Pod type the desired machine must be located in.")
        )
        parser.add_argument(
            "--subnet",
            nargs="*",
            help=("Subnet(s) the desired machine must be linked to."),
        )
        parser.add_argument(
            "--tag", nargs="*", help=("Tags the desired machine must match.")
        )
        parser.add_argument(
            "--zone", help=("Zone the desired machine must be located in.")
        )
        parser.add_argument(
            "--not-fabric",
            nargs="*",
            help=("Fabric(s) the desired machine must NOT be connected to."),
        )
        parser.add_argument(
            "--not-subnet",
            nargs="*",
            help=("Subnets(s) the desired machine must NOT be linked to."),
        )
        parser.add_argument(
            "--not-tag", nargs="*", help=("Tags the desired machine must NOT match.")
        )
        parser.add_argument(
            "--not-zone",
            nargs="*",
            help=("Zone(s) the desired machine must NOT belong in."),
        )
        parser.other.add_argument(
            "--agent-name", help=("Agent name to attach to the acquire machine.")
        )
        if with_comment:
            parser.other.add_argument(
                "--comment", help=("Reason for allocating the machine.")
            )
        parser.other.add_argument(
            "--bridge-all",
            action="store_true",
            default=None,
            help=(
                "Automatically create a bridge on all interfaces on the "
                "allocated machine."
            ),
        )
        parser.other.add_argument(
            "--bridge-stp",
            action="store_true",
            default=None,
            help=(
                "Turn spaning tree protocol on or off for the bridges created "
                "with --bridge-all."
            ),
        )
        parser.other.add_argument(
            "--bridge-fd",
            type=int,
            help=(
                "Set the forward delay in seconds on the bridges created with "
                "--bridge-all."
            ),
        )
        if with_dry_run:
            parser.other.add_argument(
                "--dry-run",
                action="store_true",
                default=None,
                help=(
                    "Don't actually acquire the machine just return the "
                    "machine that would have been acquired."
                ),
            )

    @asynchronous
    async def allocate(self, origin, options):
        if options.hostname:
            me = await origin.Users.whoami()
            machines = await origin.Machines.read(hostnames=[options.hostname])
            if len(machines) == 0:
                raise CommandError("Unable to find machine %s." % options.hostname)
            machine = machines[0]
            if (
                machine.status == NodeStatus.ALLOCATED
                and machine.owner.username == me.username
            ):
                return False, machine
            elif machine.status != NodeStatus.READY:
                raise CommandError("Unable to allocate machine %s." % options.hostname)
        params = utils.remove_None(
            {
                "hostname": options.hostname,
                "architectures": options.arch,
                "cpus": options.cpus,
                "memory": options.memory,
                "fabrics": options.fabric,
                "interfaces": options.interface,
                "pod": options.pod,
                "pod_type": options.pod_type,
                "subnets": options.subnet,
                "tags": options.tag,
                "not_fabrics": options.not_fabric,
                "not_subnets": options.not_subnet,
                "not_zones": options.not_zone,
                "agent_name": options.agent_name,
                "comment": options.comment,
                "bridge_all": options.bridge_all,
                "bridge_stp": options.bridge_stp,
                "bridge_fd": options.bridge_fd,
                "dry_run": getattr(options, "dry_run", False),
            }
        )
        machine = await origin.Machines.allocate(**params)
        if options.hostname and machine.hostname != options.hostname:
            await machine.release()
            raise CommandError(
                "MAAS failed to allocate machine %s; "
                "instead it allocated %s." % (options.hostname, machine.hostname)
            )
        return True, machine

    def execute(self, origin, options):
        with utils.Spinner() as context:
            context.msg = colorized("{automagenta}Allocating{/automagenta}")
            _, machine = self.allocate(origin, options)
        print(colorized("{autoblue}Allocated{/autoblue} %s") % machine.hostname)


class MachineWorkMixin:
    """Mixin that helps with performing actions across a set of machinse."""

    @asynchronous
    async def _async_perform_action(
        self, context, action, machines, params, progress_title, success_title
    ):
        def _update_msg(remaining):
            """Update the spinner message."""
            if len(remaining) == 1:
                msg = remaining[0].hostname
            elif len(remaining) == 2:
                msg = "%s and %s" % (remaining[0].hostname, remaining[1].hostname)
            else:
                msg = "%s machines" % len(remaining)
            context.msg = colorized(
                "{autoblue}%s{/autoblue} %s" % (progress_title, msg)
            )

        async def _perform(machine, params, remaining):
            """Updates the messages as actions complete."""
            try:
                await getattr(machine, action)(**params)
            except Exception as exc:
                remaining.remove(machine)
                _update_msg(remaining)
                context.print(colorized("{autored}Error:{/autored} %s") % str(exc))
                raise
            else:
                remaining.remove(machine)
                _update_msg(remaining)
                context.print(
                    colorized("{autogreen}%s{/autogreen} %s")
                    % (success_title, machine.hostname)
                )

        _update_msg(machines)
        results = await asyncio.gather(
            *[_perform(machine, params, machines) for machine in machines],
            return_exceptions=True
        )
        failures = [result for result in results if isinstance(result, Exception)]
        if len(failures) > 0:
            return 1
        return 0

    def perform_action(self, action, machines, params, progress_title, success_title):
        """Perform the action on the set of machines."""
        if len(machines) == 0:
            return 0
        with utils.Spinner() as context:
            return self._async_perform_action(
                context, action, list(machines), params, progress_title, success_title
            )

    def get_machines(self, origin, hostnames):
        """Return a set of machines based on `hostnames`.

        Any hostname that is not found will result in an error.
        """
        hostnames = {hostname: True for hostname in hostnames}
        machines = origin.Machines.read(hostnames=hostnames)
        machines = [
            machine for machine in machines if hostnames.pop(machine.hostname, False)
        ]
        if len(hostnames) > 0:
            raise CommandError(
                "Unable to find %s %s."
                % ("machines" if len(hostnames) > 1 else "machine", ",".join(hostnames))
            )
        return machines


class MachineSSHMixin:
    """Mixin that provides the ability to SSH."""

    def add_ssh_options(self, parser):
        """Add the SSH arguments to the `parser`."""
        parser.add_argument(
            "--username", metavar="USER", help=("Username for the SSH connection.")
        )
        parser.add_argument(
            "--boot-only",
            action="store_true",
            help=("Only use the IP addresses on the machine's boot interface."),
        )

    def get_ip_addresses(self, machine, *, boot_only=False, discovered=False):
        """Return all IP address for `machine`.

        IP address from `boot_interface` come first.
        """
        boot_ips = [
            link.ip_address for link in machine.boot_interface.links if link.ip_address
        ]
        if boot_only:
            if boot_ips:
                return boot_ips
            elif discovered:
                return [
                    link.ip_address
                    for link in machine.boot_interface.discovered
                    if link.ip_address
                ]
            else:
                return []
        else:
            other_ips = [
                link.ip_address
                for interface in machine.interfaces
                for link in interface.links
                if (interface.id != machine.boot_interface.id and link.ip_address)
            ]
            ips = boot_ips + other_ips
            if ips:
                return ips
            elif discovered:
                return [
                    link.ip_address
                    for link in machine.boot_interface.discovered
                    if link.ip_address
                ] + [
                    link.ip_address
                    for interface in machine.interfaces
                    for link in interface.discovered
                    if (interface.id != machine.boot_interface.id and link.ip_address)
                ]
            else:
                return []

    @asynchronous
    async def _async_get_sshable_ips(self, ip_addresses):
        """Return list of all IP address that could be pinged."""

        async def _async_ping(ip_address):
            try:
                reader, writer = await asyncio.wait_for(
                    asyncio.open_connection(ip_address, 22), timeout=5
                )
            except (OSError, TimeoutError):
                return None
            try:
                line = await reader.readline()
            finally:
                writer.close()
            if line.startswith(b"SSH-"):
                return ip_address

        ssh_ips = await asyncio.gather(
            *[_async_ping(ip_address) for ip_address in ip_addresses]
        )
        return [ip_address for ip_address in ssh_ips if ip_address is not None]

    def _check_ssh(self, *args):
        """Check if SSH connection can be made to IP with username."""
        ssh = subprocess.Popen(
            args,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        ssh.wait()
        return ssh.returncode == 0

    def _determine_username(self, ip):
        """SSH in as root and determine the username."""
        ssh = subprocess.Popen(
            [
                "ssh",
                "-o",
                "UserKnownHostsFile=/dev/null",
                "-o",
                "StrictHostKeyChecking=no",
                "root@%s" % ip,
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        first_line = ssh.stdout.readline()
        ssh.kill()
        ssh.wait()
        if first_line:
            match = re.search(
                r"Please login as the user \"(\w+)\" rather than "
                r"the user \"root\".",
                first_line.decode("utf-8"),
            )
            if match:
                return match.groups()[0]
        else:
            return None

    def ssh(
        self,
        machine,
        *,
        username=None,
        command=None,
        boot_only=False,
        discovered=False,
        wait=300
    ):
        """SSH into `machine`."""
        start_time = time.monotonic()
        with utils.Spinner() as context:
            context.msg = colorized(
                "{autoblue}Determining{/autoblue} best IP for %s" % (machine.hostname)
            )
            ip_addresses = self.get_ip_addresses(
                machine, boot_only=boot_only, discovered=discovered
            )
            if len(ip_addresses) > 0:
                pingable_ips = self._async_get_sshable_ips(ip_addresses)
                while len(pingable_ips) == 0 and (time.monotonic() - start_time) < wait:
                    time.sleep(5)
                    pingable_ips = self._async_get_sshable_ips(ip_addresses)
                if len(pingable_ips) == 0:
                    raise CommandError(
                        "No IP addresses on %s can be reached." % (machine.hostname)
                    )
                else:
                    ip = pingable_ips[0]
            else:
                raise CommandError("%s has no IP addresses." % machine.hostname)

            if username is None:
                context.msg = colorized(
                    "{autoblue}Determining{/autoblue} SSH username on %s"
                    % (machine.hostname)
                )
                username = self._determine_username(ip)
                while username is None and (time.monotonic() - start_time) < wait:
                    username = self._determine_username(ip)
                if username is None:
                    raise CommandError("Failed to determine the username for SSH.")

            conn_str = "%s@%s" % (username, ip)
            args = [
                "ssh",
                "-o",
                "UserKnownHostsFile=/dev/null",
                "-o",
                "StrictHostKeyChecking=no",
                conn_str,
            ]

            context.msg = colorized(
                "{automagenta}Waiting{/automagenta} for SSH on %s" % (machine.hostname)
            )
            check_args = args + ["echo"]
            connectable = self._check_ssh(*check_args)
            while not connectable and (time.monotonic() - start_time) < wait:
                time.sleep(5)
                connectable = self._check_ssh(*check_args)
            if not connectable:
                raise CommandError(
                    "SSH never started on %s using IP %s." % (machine.hostname, ip)
                )

        if command is not None:
            args.append(command)
        ssh = subprocess.Popen(
            args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr
        )
        ssh.wait()
        return ssh.returncode


class MachineReleaseMixin(MachineWorkMixin):
    """Mixin that provide releasing machines."""

    def add_release_options(self, parser):
        parser.add_argument(
            "--erase", action="store_true", help=("Erase the disk when releasing.")
        )
        parser.add_argument(
            "--secure-erase",
            action="store_true",
            help=("Use the drives secure erase feature if available on the disk."),
        )
        parser.add_argument(
            "--quick-erase",
            action="store_true",
            help=(
                "Wipe the just the beginning and end of the disk. "
                "This is not secure."
            ),
        )

    def get_release_params(self, options):
        return utils.remove_None(
            {
                "erase": options.erase,
                "secure_erase": options.secure_erase,
                "quick_erase": options.quick_erase,
            }
        )

    def release(self, machines, params):
        wait = params.get("wait", True)
        return self.perform_action(
            "release",
            machines,
            params,
            "Releasing",
            "Released" if wait else "Releasing",
        )


class cmd_deploy(cmd_allocate, MachineSSHMixin, MachineReleaseMixin):
    """Allocate and deploy machine.

    See `help allocate` for more details on the allocation parameters.
    """

    def __init__(self, parser):
        super(cmd_deploy, self).__init__(
            parser, with_hostname=False, with_comment=False, with_dry_run=False
        )
        parser.add_argument(
            "image",
            nargs="?",
            help=(
                "Image to deploy to the machine (e.g. ubuntu/xenial or " "just xenial)."
            ),
        )
        parser.add_argument("hostname", nargs="?", help=("Hostname of the machine."))
        parser.add_argument(
            "--hwe-kernel",
            help=(
                "Hardware enablement kernel to use with the image. Only used "
                "when deploying Ubuntu."
            ),
        )
        parser.add_argument(
            "--user-data",
            metavar="FILE",
            type=lambda arg: validate_file(parser, arg),
            help=("User data that gets run on the machine once it has " "deployed."),
        )
        parser.add_argument(
            "--b64-user-data",
            metavar="BASE64",
            help=(
                "Base64 encoded string of the user data that gets run on the "
                "machine once it has deployed."
            ),
        )
        parser.add_argument(
            "--ssh",
            action="store_true",
            help=("SSH into the machine once its deployed."),
        )
        self.add_ssh_options(parser)
        parser.add_argument(
            "--release-on-exit",
            action="store_true",
            help=(
                "Release the machine once the SSH connection is closed. "
                "Only used with --ssh is provided."
            ),
        )
        self.add_release_options(parser)
        parser.other.add_argument(
            "--comment", help=("Reason for deploying the machine.")
        )
        parser.other.add_argument(
            "--no-wait",
            action="store_true",
            help=("Don't wait for the deploy to complete."),
        )
        parser.other.add_argument(
            "--install-kvm", action="store_true", help=("Install KVM on machine")
        )

    def _get_deploy_options(self, options):
        """Return the deployment options based on command line."""
        user_data = None
        if options.user_data and options.b64_user_data:
            raise CommandError("Cannot provide both --user-data and --b64-user-data.")
        if options.b64_user_data:
            user_data = options.b64_user_data
        if options.user_data:
            user_data = base64_file(options.user_data).decode("ascii")
        return utils.remove_None(
            {
                "distro_series": options.image,
                "hwe_kernel": options.hwe_kernel,
                "user_data": user_data,
                "comment": options.comment,
                "install_kvm": options.install_kvm,
                "wait": False,
            }
        )

    def _handle_abort(self, machine, allocated):
        """Handle the user aborting mid deployment."""
        abort = yes_or_no("Abort deployment?")
        if abort:
            with utils.Spinner() as context:
                if allocated:
                    context.msg = colorized("{autoblue}Releasing{/autoblue} %s") % (
                        machine.hostname
                    )
                    machine.release()
                    context.print(
                        colorized("{autoblue}Released{/autoblue} %s")
                        % (machine.hostname)
                    )
                else:
                    context.msg = colorized("{autoblue}Aborting{/autoblue} %s") % (
                        machine.hostname
                    )
                    machine.abort()
                    context.print(
                        colorized("{autoblue}Aborted{/autoblue} %s")
                        % (machine.hostname)
                    )

    def execute(self, origin, options):
        deploy_options = self._get_deploy_options(options)
        allocated, machine = False, None
        try:
            with utils.Spinner() as context:
                if options.hostname:
                    context.msg = colorized("{autoblue}Allocating{/autoblue} %s") % (
                        options.hostname
                    )
                else:
                    context.msg = colorized("{autoblue}Searching{/autoblue}")
                allocated, machine = self.allocate(origin, options)
                context.msg = (
                    colorized("{autoblue}Deploying{/autoblue} %s") % machine.hostname
                )
                try:
                    machine = machine.deploy(**deploy_options)
                except CallError:
                    if allocated:
                        machine.release()
                    raise
                if not options.no_wait:
                    context.msg = colorized(
                        "{autoblue}Deploying{/autoblue} %s on %s"
                    ) % (machine.distro_series, machine.hostname)
                    while machine.status == NodeStatus.DEPLOYING:
                        time.sleep(15)
                        machine.refresh()
                        context.msg = colorized(
                            "{autoblue}Deploying{/autoblue} %s on %s: %s"
                        ) % (
                            machine.distro_series,
                            machine.hostname,
                            machine.status_message,
                        )
        except KeyboardInterrupt:
            if sys.stdout.isatty() and machine is not None:
                self._handle_abort(machine, allocated)
            raise

        if machine.status == NodeStatus.FAILED_DEPLOYMENT:
            raise CommandError(
                "Deployment of %s on %s failed."
                % (machine.distro_series, machine.hostname)
            )
        elif machine.status == NodeStatus.DEPLOYED:
            print(
                colorized("{autoblue}Deployed{/autoblue} %s on %s")
                % (machine.distro_series, machine.hostname)
            )
        elif machine.status == NodeStatus.DEPLOYING:
            print(
                colorized("{autoblue}Deploying{/autoblue} %s on %s")
                % (machine.distro_series, machine.hostname)
            )
        else:
            raise CommandError(
                "Machine %s transitioned to an unexpected state of %s."
                % (machine.hostname, machine.status_name)
            )

        if options.ssh:
            machine.refresh()
            code = self.ssh(
                machine, username=options.username, boot_only=options.boot_only
            )
            if code == 0 and options.release_on_exit:
                release_params = self.get_release_params(options)
                release_params["wait"] = True
                self.release([machine], release_params)


class cmd_commission(OriginCommand, MachineSSHMixin, MachineWorkMixin):
    """Commission machine."""

    def __init__(self, parser):
        super(cmd_commission, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="*", help=("Hostname of the machine to commission.")
        )
        parser.add_argument(
            "--all",
            action="store_true",
            help=("Commission all machines that can be commissioned."),
        )
        parser.add_argument(
            "--new", action="store_true", help=("Commission all new machines.")
        )
        parser.add_argument(
            "--skip-networking",
            action="store_true",
            help=(
                "Skip machine network discovery, keeping the current interface "
                "configuration for the machine."
            ),
        )
        parser.add_argument(
            "--skip-storage",
            action="store_true",
            help=(
                "Skip machine storage discovery, keeping the current storage "
                "configuration for the machine."
            ),
        )
        parser.add_argument(
            "--scripts",
            nargs="*",
            metavar="SCRIPT",
            help=("Run only the selected commissioning scripts."),
        )
        parser.add_argument(
            "--ssh",
            action="store_true",
            help=("SSH into the machine during commissioning."),
        )
        self.add_ssh_options(parser)
        parser.other.add_argument(
            "--no-wait",
            action="store_true",
            help=("Don't wait for the commisisoning to complete."),
        )

    def execute(self, origin, options):
        if options.hostname and options.all:
            raise CommandError("Cannot pass both hostname and --all.")
        if options.hostname and options.new:
            raise CommandError("Cannot pass both hostname and --new.")
        if not options.hostname and not options.all and not options.new:
            raise CommandError("Missing parameter hostname, --all, or --new.")
        if options.ssh and (len(options.hostname) > 1 or options.all or options.new):
            raise CommandError("--ssh can only be used when commissioning one machine.")
        if options.all:
            machines = origin.Machines.read()
            machines = [
                machine
                for machine in machines
                if machine.status
                in [NodeStatus.NEW, NodeStatus.READY, NodeStatus.FAILED_COMMISSIONING]
            ]
        elif options.new:
            machines = origin.Machines.read()
            machines = [
                machine for machine in machines if machine.status == NodeStatus.NEW
            ]
        else:
            machines = self.get_machines(origin, options.hostname)
        params = utils.remove_None(
            {
                "enable_ssh": options.ssh,
                "skip_networking": options.skip_networking,
                "skip_storage": options.skip_storage,
                "commissioning_scripts": options.scripts,
                "wait": False if options.no_wait else True,
            }
        )
        try:
            rc = self.perform_action(
                "commission",
                machines,
                params,
                "Commissioning",
                "Commissioning" if options.no_wait else "Commissioned",
            )
        except KeyboardInterrupt:
            if sys.stdout.isatty():
                abort = yes_or_no("Abort commissioning?")
                if abort:
                    return self.perform_action(
                        "abort", machines, {}, "Aborting", "Aborted"
                    )
                else:
                    return 1
        if rc == 0 and len(machines) > 0 and options.ssh:
            machine = machines[0]
            machine.refresh()
            rc = self.ssh(
                machine,
                username=options.username,
                boot_only=options.boot_only,
                discovered=True,
            )
            if rc == 0:
                return self.perform_action(
                    "power_off", [machine], {}, "Powering off", "Powered off"
                )
        return rc


class cmd_release(OriginCommand, MachineReleaseMixin):
    """Release machine."""

    def __init__(self, parser):
        super(cmd_release, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="*", help=("Hostname of the machine to release.")
        )
        parser.add_argument(
            "--all", action="store_true", help=("Release all machines owned by you.")
        )
        parser.add_argument("--comment", help=("Reason for releasing the machine."))
        self.add_release_options(parser)
        parser.other.add_argument(
            "--no-wait",
            action="store_true",
            help=("Don't wait for the release to complete."),
        )

    def execute(self, origin, options):
        if options.hostname and options.all:
            raise CommandError("Cannot pass both hostname and --all.")
        if not options.hostname and not options.all:
            raise CommandError("Missing parameter hostname or --all.")
        params = self.get_release_params(options)
        params["wait"] = False if options.no_wait else True
        if options.all:
            me = origin.Users.whoami()
            machines = origin.Machines.read()
            machines = [
                machine
                for machine in machines
                if (
                    machine.owner is not None
                    and machine.owner.username == me.username
                    and (
                        machine.status
                        not in [NodeStatus.COMMISSIONING, NodeStatus.TESTING]
                    )
                )
            ]
        else:
            machines = self.get_machines(origin, options.hostname)
        return self.release(machines, params)


class cmd_abort(OriginCommand, MachineWorkMixin):
    """Abort machine's current action."""

    def __init__(self, parser):
        super(cmd_abort, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="+", help=("Hostname of the machine to abort the action.")
        )
        parser.add_argument("--comment", help=("Reason for aborting the action."))

    def execute(self, origin, options):
        params = utils.remove_None({"comment": options.comment})
        machines = self.get_machines(origin, options.hostname)
        return self.perform_action("abort", machines, params, "Aborting", "Aborted")


class cmd_mark_fixed(OriginCommand, MachineWorkMixin):
    """Mark machine fixed."""

    def __init__(self, parser):
        super(cmd_mark_fixed, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="+", help=("Hostname of the machine to mark fixed.")
        )
        parser.add_argument("--comment", help=("Reason for marking the machine fixed."))

    def execute(self, origin, options):
        machines = self.get_machines(origin, options.hostname)
        return self.perform_action(
            "mark_fixed", machines, {}, "Marking fixed", "Marked fixed"
        )


class cmd_mark_broken(OriginCommand, MachineWorkMixin):
    """Mark machine broken."""

    def __init__(self, parser):
        super(cmd_mark_broken, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="+", help=("Hostname of the machine to mark broken.")
        )
        parser.add_argument(
            "--comment", help=("Reason for marking the machine broken.")
        )

    def execute(self, origin, options):
        machines = self.get_machines(origin, options.hostname)
        return self.perform_action(
            "mark_broken", machines, {}, "Marking broken", "Marked broken"
        )


class cmd_ssh(OriginCommand, MachineWorkMixin, MachineSSHMixin):
    """SSH into a machine."""

    def __init__(self, parser):
        super(cmd_ssh, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs=1, help=("Hostname of the machine to SSH to.")
        )
        parser.add_argument(
            "command",
            nargs="?",
            default=None,
            help=("Hostname of the machine to SSH to."),
        )
        self.add_ssh_options(parser)

    def execute(self, origin, options):
        machine = self.get_machines(origin, options.hostname)[0]
        return self.ssh(
            machine,
            username=options.username,
            command=options.command,
            boot_only=options.boot_only,
        )


class cmd_power_on(OriginCommand, MachineWorkMixin):
    """Power on machine."""

    def __init__(self, parser):
        super(cmd_power_on, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="+", help=("Hostname of the machine to power on.")
        )
        parser.add_argument("--comment", help=("Reason for powering the machine on."))

    def execute(self, origin, options):
        machines = self.get_machines(origin, options.hostname)
        return self.perform_action(
            "power_on", machines, {}, "Powering on", "Powered on"
        )


class cmd_power_off(OriginCommand, MachineWorkMixin):
    """Power off machine."""

    def __init__(self, parser):
        super(cmd_power_off, self).__init__(parser)
        parser.add_argument(
            "hostname", nargs="+", help=("Hostname of the machine to power off.")
        )
        parser.add_argument("--comment", help=("Reason for powering the machine off."))

    def execute(self, origin, options):
        machines = self.get_machines(origin, options.hostname)
        return self.perform_action(
            "power_off", machines, {}, "Powering off", "Powered off"
        )


def register(parser):
    """Register commands with the given parser."""
    cmd_machines.register(parser)
    cmd_machine.register(parser)
    cmd_allocate.register(parser)
    cmd_deploy.register(parser)
    cmd_commission.register(parser)
    cmd_release.register(parser)
    cmd_abort.register(parser)
    cmd_mark_fixed.register(parser)
    cmd_mark_broken.register(parser)
    cmd_power_off.register(parser)
    cmd_power_on.register(parser)
    cmd_ssh.register(parser)



"""Objects for boot resources."""

__all__ = ["BootResource", "BootResources"]

import enum
import hashlib
import io
from urllib.parse import urlparse

import aiohttp

from . import (
    check,
    check_optional,
    mapping_of,
    Object,
    ObjectField,
    ObjectSet,
    ObjectType,
)
from .. import utils
from ..bones import CallError


def calc_size_and_sha265(content: io.IOBase, chunk_size: int):
    """Calculates the size and the sha2566 value of the content."""
    size = 0
    sha256 = hashlib.sha256()
    content.seek(0, io.SEEK_SET)
    while True:
        buf = content.read(chunk_size)
        length = len(buf)
        size += length
        sha256.update(buf)
        if length != chunk_size:
            break
    return size, sha256.hexdigest()


class BootResourceFileType(enum.Enum):

    TGZ = "tgz"
    DDTGZ = "ddtgz"


class BootResourceFile(Object):
    """A boot resource file."""

    filename = ObjectField.Checked("filename", check(str), readonly=True)
    filetype = ObjectField.Checked("filetype", check(str), readonly=True)
    size = ObjectField.Checked("size", check(int), readonly=True)
    sha256 = ObjectField.Checked("sha256", check(str), readonly=True)
    complete = ObjectField.Checked("complete", check(bool), readonly=True)


class BootResourceSet(Object):
    """A boot resource set."""

    version = ObjectField.Checked("version", check(str), readonly=True)
    size = ObjectField.Checked("size", check(int), readonly=True)
    label = ObjectField.Checked("label", check(str), readonly=True)
    complete = ObjectField.Checked("complete", check(bool), readonly=True)
    files = ObjectField.Checked(
        "files", mapping_of(BootResourceFile), default=None, readonly=True
    )


class BootResourcesType(ObjectType):
    """Metaclass for `BootResources`."""

    async def read(cls):
        """Get list of `BootResource`'s.

        Each `BootResource` from the API is limited in that it does not
        contain the `sets` in the output. To get full data for each object
        a read needs to be performed on each object::

            resources = BootResources.read()
            resources = [
                BootResource.read(resource.id)
                for resource in resources
            ]

        """
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def start_import(cls):
        """Start the import of `BootResource`'s."""
        # Cannot use cls._handler.import() as import is a reserved statement.
        return getattr(cls._handler, "import")()

    async def stop_import(cls):
        """Stop the import of `BootResource`'s."""
        return cls._handler.stop_import()

    async def create(
        cls,
        name: str,
        architecture: str,
        content: io.IOBase,
        *,
        title: str = "",
        filetype: BootResourceFileType = BootResourceFileType.TGZ,
        chunk_size=(1 << 22),
        progress_callback=None
    ):
        """Create a `BootResource`.

        Creates an uploaded boot resource with `content`. The `content` is
        uploaded in chunks of `chunk_size`. `content` must be seekable as the
        first pass through the `content` will calculate the size and sha256
        value then the second pass will perform the actual upload.

        :param name: Name of the boot resource. Must be in format 'os/release'.
        :type name: `str`
        :param architecture: Architecture of the boot resource. Must be in
            format 'arch/subarch'.
        :type architecture: `str`
        :param content: Content of the boot resource.
        :type content: `io.IOBase`
        :param title: Title of the boot resource.
        :type title: `str`
        :param filetype: Type of file in content.
        :type filetype: `str`
        :param chunk_size: Size in bytes to upload to MAAS in chunks.
            (Default is 4 MiB).
        :type chunk_size: `int`
        :param progress_callback: Called to inform the current progress of the
            upload. One argument is passed with the progress as a precentage.
            If the resource was already complete and no content
            needed to be uploaded then this callback will never be called.
        :type progress_callback: Callable
        :returns: Create boot resource.
        :rtype: `BootResource`.
        """
        if "/" not in name:
            raise ValueError("name must be in format os/release; missing '/'")
        if "/" not in architecture:
            raise ValueError("architecture must be in format arch/subarch; missing '/'")
        if not content.readable():
            raise ValueError("content must be readable")
        elif not content.seekable():
            raise ValueError("content must be seekable")
        if chunk_size <= 0:
            raise ValueError("chunk_size must be greater than 0, not %d" % chunk_size)

        size, sha256 = calc_size_and_sha265(content, chunk_size)
        resource = cls._object(
            await cls._handler.create(
                name=name,
                architecture=architecture,
                title=title,
                filetype=filetype.value,
                size=str(size),
                sha256=sha256,
            )
        )
        newest_set = max(resource.sets, default=None)
        assert newest_set is not None
        resource_set = resource.sets[newest_set]
        assert len(resource_set.files) == 1
        rfile = list(resource_set.files.values())[0]
        if rfile.complete:
            # Already created and fully up-to-date.
            return resource
        else:
            # Upload in chunks and reload boot resource.
            await cls._upload_chunks(rfile, content, chunk_size, progress_callback)
            return cls._object.read(resource.id)

    async def _upload_chunks(
        cls,
        rfile: BootResourceFile,
        content: io.IOBase,
        chunk_size: int,
        progress_callback=None,
    ):
        """Upload the `content` to `rfile` in chunks using `chunk_size`."""
        content.seek(0, io.SEEK_SET)
        upload_uri = (
            urlparse(cls._handler.uri)._replace(path=rfile._data["upload_uri"]).geturl()
        )
        uploaded_size = 0

        insecure = cls._handler.session.insecure
        connector = aiohttp.TCPConnector(verify_ssl=(not insecure))
        session = aiohttp.ClientSession(connector=connector)

        async with session:
            while True:
                buf = content.read(chunk_size)
                length = len(buf)
                if length > 0:
                    uploaded_size += length
                    await cls._put_chunk(session, upload_uri, buf)
                    if progress_callback is not None:
                        progress_callback(uploaded_size / rfile.size)
                if length != chunk_size:
                    break

    async def _put_chunk(
        cls, session: aiohttp.ClientSession, upload_uri: str, buf: bytes
    ):
        """Upload one chunk to `upload_uri`."""
        # Build the correct headers.
        headers = {
            "Content-Type": "application/octet-stream",
            "Content-Length": "%s" % len(buf),
        }
        credentials = cls._handler.session.credentials
        if credentials is not None:
            utils.sign(upload_uri, headers, credentials)

        # Perform upload of chunk.
        async with await session.put(upload_uri, data=buf, headers=headers) as response:
            if response.status != 200:
                content = await response.read()
                request = {
                    "body": buf,
                    "headers": headers,
                    "method": "PUT",
                    "uri": upload_uri,
                }
                raise CallError(request, response, content, None)


class BootResources(ObjectSet, metaclass=BootResourcesType):
    """The set of boot resources."""


class BootResourceType(ObjectType):
    async def read(cls, id: int):
        """Get `BootResource` by `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class BootResource(Object, metaclass=BootResourceType):
    """A boot resource."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    type = ObjectField.Checked("type", check(str), check(str), readonly=True)
    name = ObjectField.Checked("name", check(str), check(str), readonly=True)
    architecture = ObjectField.Checked(
        "architecture", check(str), check(str), readonly=True
    )
    subarches = ObjectField.Checked(
        "subarches",
        check_optional(str),
        check_optional(str),
        default=None,
        readonly=True,
    )
    sets = ObjectField.Checked(
        "sets", mapping_of(BootResourceSet), default=None, readonly=True
    )

    def __repr__(self):
        return super(BootResource, self).__repr__(
            fields={"type", "name", "architecture"}
        )

    async def delete(self):
        """Delete boot resource."""
        await self._handler.delete(id=self.id)



""" Custom errors for libmaas """

__all__ = ["MAASException", "OperationNotAllowed"]


class MAASException(Exception):
    def __init__(self, msg, obj):
        super().__init__(msg)
        self.obj = obj


class OperationNotAllowed(Exception):
    """MAAS says this operation cannot be performed."""


class ObjectNotLoaded(Exception):
    """Object is not loaded."""


class CannotDelete(Exception):
    """Object cannot be deleted."""


class PowerError(MAASException):
    """Machine failed to power on or off."""



"""Objects for fabrics."""

__all__ = ["Fabrics", "Fabric"]

from . import check, Object, ObjectField, ObjectFieldRelatedSet, ObjectSet, ObjectType
from ..errors import CannotDelete


class FabricsType(ObjectType):
    """Metaclass for `Fabrics`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(
        cls, *, name: str = None, description: str = None, class_type: str = None
    ):
        """
        Create a `Fabric` in MAAS.

        :param name: The name of the `Fabric` (optional, will be given a
        default value if not specified).
        :type name: `str`
        :param description: A description of the `Fabric` (optional).
        :type description: `str`
        :param class_type: The class type of the `Fabric` (optional).
        :type class_type: `str`
        :returns: The created Fabric
        :rtype: `Fabric`
        """
        params = {}
        if name is not None:
            params["name"] = name
        if description is not None:
            params["description"] = description
        if class_type is not None:
            params["class_type"] = class_type
        return cls._object(await cls._handler.create(**params))


class Fabrics(ObjectSet, metaclass=FabricsType):
    """The set of Fabrics stored in MAAS."""


class FabricType(ObjectType):
    """Metaclass for `Fabric`."""

    _default_fabric_id = 0

    async def get_default(cls):
        """
        Get the 'default' Fabric for the MAAS.
        """
        data = await cls._handler.read(id=cls._default_fabric_id)
        return cls(data)

    async def read(cls, id: int):
        """Get a `Fabric` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class Fabric(Object, metaclass=FabricType):
    """A Fabric."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    name = ObjectField.Checked("name", check(str), check(str))
    vlans = ObjectFieldRelatedSet("vlans", "Vlans")

    async def delete(self):
        """Delete this Fabric."""
        if self.id == self._origin.Fabric._default_fabric_id:
            raise CannotDelete("Default fabric cannot be deleted.")
        await self._handler.delete(id=self.id)



"""Objects for subnets."""

__all__ = ["Subnets", "Subnet"]

from typing import Sequence, Union

from . import (
    check,
    check_optional,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectSet,
    ObjectType,
    to,
)
from .vlans import Vlan
from ..enum import RDNSMode


class SubnetsType(ObjectType):
    """Metaclass for `Subnets`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(
        cls,
        cidr: str,
        vlan: Union[Vlan, int] = None,
        *,
        name: str = None,
        description: str = None,
        gateway_ip: str = None,
        rdns_mode: RDNSMode = None,
        dns_servers: Union[Sequence[str], str] = None,
        managed: bool = None
    ):
        """
        Create a `Subnet` in MAAS.

        :param cidr: The cidr of the `Subnet` (required).
        :type cidr: `str`
        :param vlan: The VLAN of the `Subnet` (required).
        :type vlan: `Vlan`
        :param name: The name of the `Subnet` (optional, will be given a
        default value if not specified).
        :type name: `str`
        :param description: A description of the `Subnet` (optional).
        :type description: `str`
        :param gateway_ip: The gateway IP address for the `Subnet` (optional).
        :type gateway_ip: `str`
        :param rdns_mode: The reverse DNS mode for the `Subnet` (optional).
        :type rdns_mode: `RDNSMode`
        :param managed: Whether the `Subnet` is managed by MAAS (optional).
        :type managed: `bool`
        :returns: The created Subnet
        :rtype: `Subnet`
        """
        params = {"cidr": cidr}

        if isinstance(vlan, int):
            params["vlan"] = vlan
        elif isinstance(vlan, Vlan):
            params["vlan"] = vlan.id
        else:
            raise TypeError("vlan must be Vlan or int, not %s" % (type(vlan).__class__))
        if name is not None:
            params["name"] = name
        if description is not None:
            params["description"] = description
        if gateway_ip is not None:
            params["gateway_ip"] = gateway_ip
        if rdns_mode is not None:
            params["rdns_mode"] = rdns_mode
        if isinstance(dns_servers, Sequence):
            if len(dns_servers) > 0:
                params["dns_servers"] = ",".join(dns_servers)
        elif dns_servers is not None:
            params["dns_servers"] = dns_servers
        if managed is not None:
            params["managed"] = managed
        return cls._object(await cls._handler.create(**params))


class Subnets(ObjectSet, metaclass=SubnetsType):
    """The set of Subnets stored in MAAS."""


class SubnetType(ObjectType):
    """Metaclass for `Subnet`"""

    async def read(cls, id: int):
        """Get a `Subnet` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class Subnet(Object, metaclass=SubnetType):
    """A Subnet."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    cidr = ObjectField.Checked("cidr", check(str), readonly=True)
    name = ObjectField.Checked("name", check(str))

    # description is allowed in the create call and displayed in the UI
    # but never returned by the API

    vlan = ObjectFieldRelated("vlan", "Vlan", use_data_setter=True)

    # This should point to the space object and not just the string.
    space = ObjectField.Checked("space", check(str))

    active_discovery = ObjectField.Checked("active_discovery", check(bool))
    allow_proxy = ObjectField.Checked("allow_proxy", check(bool))
    managed = ObjectField.Checked("managed", check(bool))
    gateway_ip = ObjectField.Checked("gateway_ip", check_optional(str))
    rdns_mode = ObjectField.Checked("rdns_mode", to(RDNSMode))
    dns_servers = ObjectField.Checked("dns_servers", check(list))

    def __repr__(self):
        return super(Subnet, self).__repr__(fields={"cidr", "name", "vlan"})

    async def save(self):
        """Save this subnet."""
        if "vlan" in self._changed_data and self._changed_data["vlan"]:
            # Update uses the ID of the VLAN, not the VLAN object.
            self._changed_data["vlan"] = self._changed_data["vlan"]["id"]
            if (
                self._orig_data["vlan"]
                and "id" in self._orig_data["vlan"]
                and self._changed_data["vlan"] == (self._orig_data["vlan"]["id"])
            ):
                # VLAN didn't really change, the object was just set to the
                # same VLAN.
                del self._changed_data["vlan"]
        await super(Subnet, self).save()

    async def delete(self):
        """Delete this Subnet."""
        await self._handler.delete(id=self.id)



"""Interact with a remote MAAS (https://maas.ubuntu.com/).

These are low-level bindings that closely mirror the shape of MAAS's Web API,
hence the name "bones".
"""

__all__ = ["CallError", "SessionAPI"]

import typing

from collections import namedtuple
from collections.abc import Iterable
import json
from urllib.parse import urlparse

import aiohttp

from . import helpers
from .. import utils
from ..utils import profiles
from ..utils.maas_async import asynchronous


class SessionError(Exception):
    """Miscellaneous session-related error."""


class SessionAPI:
    """Represents an API session with a remote MAAS installation."""

    @classmethod
    @asynchronous
    async def fromURL(cls, url, *, credentials=None, insecure=False):
        """Return a `SessionAPI` for a given MAAS instance."""
        try:
            description = await helpers.fetch_api_description(url, insecure=insecure)
        except helpers.RemoteError as error:
            # For now just re-raise as SessionError.
            raise SessionError(str(error))
        else:
            session = cls(description, credentials)
            session.scheme = urlparse(url).scheme
            session.insecure = insecure
            return session

    @classmethod
    def fromProfile(cls, profile):
        """Return a `SessionAPI` from a given configuration profile.

        :see: `ProfileStore`.
        """
        session = cls(profile.description, profile.credentials)
        session.scheme = urlparse(profile.url).scheme
        session.insecure = profile.other.get("insecure", False)
        return session

    @classmethod
    def fromProfileName(cls, name):
        """Return a `SessionAPI` from a given configuration profile name.

        :see: `ProfileStore`.
        """
        with profiles.ProfileStore.open() as config:
            return cls.fromProfile(config.load(name))

    @classmethod
    @asynchronous
    async def login(cls, url, *, username=None, password=None, insecure=False):
        """Make a `SessionAPI` by logging-in with a username and password.

        :return: A tuple of ``profile`` and ``session``, where the former is
            an unsaved `Profile` instance, and the latter is a `SessionAPI`
            instance made using the profile.
        """
        profile = await helpers.login(
            url=url, username=username, password=password, insecure=insecure
        )
        session = cls(profile.description, profile.credentials)
        session.scheme = urlparse(url).scheme
        session.insecure = insecure
        return profile, session

    @classmethod
    @asynchronous
    async def connect(cls, url, *, apikey=None, insecure=False):
        """Make a `SessionAPI` by connecting with an apikey.

        :return: A tuple of ``profile`` and ``session``, where the former is
            an unsaved `Profile` instance, and the latter is a `SessionAPI`
            instance made using the profile.
        """
        profile = await helpers.connect(url=url, apikey=apikey, insecure=insecure)
        session = cls(profile.description, profile.credentials)
        session.scheme = urlparse(url).scheme
        session.insecure = insecure
        return profile, session

    # Set these on instances.
    scheme = "http"
    insecure = False
    debug = False

    def __init__(self, description, credentials=None):
        """Construct a `SessionAPI`.

        :param description: The description of the remote API. See `fromURL`.
        :param credentials: Credentials for the remote system. Optional.
        """
        super(SessionAPI, self).__init__()
        self.__description = description
        self.__credentials = credentials
        self.__populate()

    def __populate(self):
        resources = self.__description["resources"]
        if self.__credentials is None:
            for resource in resources:
                if resource["anon"] is not None:
                    handler = HandlerAPI(resource["anon"], resource, self)
                    setattr(self, handler.name, handler)
        else:
            for resource in resources:
                if resource["auth"] is not None:
                    handler = HandlerAPI(resource["auth"], resource, self)
                    setattr(self, handler.name, handler)
                elif resource["anon"] is not None:
                    handler = HandlerAPI(resource["anon"], resource, self)
                    setattr(self, handler.name, handler)

    @property
    def is_anonymous(self):
        return self.__credentials is None

    @property
    def credentials(self):
        return self.__credentials

    @property
    def description(self):
        return self.__description

    @property
    def handlers(self):
        for name, value in vars(self).items():
            if not name.startswith("_") and isinstance(value, HandlerAPI):
                yield name, value


class HandlerAPI:
    """Represents remote objects and operations, and collections thereof.

    For example, this may represent the set of all nodes and the
    operations/actions available for that set, or a single node and relevant
    operations.
    """

    def __init__(self, handler, resource, session):
        """Construct a `HandlerAPI`.

        :param handler: The handler description from the overall API
            description document. See `SessionAPI`.
        :param resource: The parent of `handler` in the API description
            document. XXX: This does not appear to be needed.
        :param session: The `SessionAPI`.
        """
        super(HandlerAPI, self).__init__()
        self.__handler = handler
        self.__resource = resource
        self.__session = session
        self.__populate()

    def __populate(self):
        self.__doc__ = self.__handler["doc"]
        actions = self.__handler["actions"]
        for action in actions:
            setattr(self, action["name"], ActionAPI(action, self))

    @property
    def name(self):
        """A stable, human-readable name and identifier for this handler."""
        return helpers.derive_resource_name(self.__handler["name"])

    @property
    def uri(self):
        """The URI for this handler.

        This will typically contain replacement patterns; these are
        interpolated in `CallAPI`.
        """
        return self.__handler["uri"]

    @property
    def params(self):
        """The set of parameters that this handler requires.

        These are the names required for interpolation into the URI.
        """
        return frozenset(self.__handler["params"])

    @property
    def path(self):
        """The path component of the URI."""
        return self.__handler["path"]

    @property
    def session(self):
        """The parent `SessionAPI`."""
        return self.__session

    @property
    def actions(self):
        return [
            (name, value)
            for name, value in vars(self).items()
            if not name.startswith("_") and isinstance(value, ActionAPI)
        ]

    def __repr__(self):
        return "<Handler %s %s>" % (self.name, self.uri)


class ActionAPI:
    """Represents a single action.

    This roughly corresponds to an HTTP verb plus a URI. Here you can bind
    parameters into the URI, as well as get information about the nature of
    the action.
    """

    def __init__(self, action, handler):
        """Construct a `ActionAPI`.

        :param action: The action description from the overall API description
            document. See `SessionAPI`.
        :param handler: The `HandlerAPI`.
        """
        super(ActionAPI, self).__init__()
        self.__action = action
        self.__handler = handler
        self.__doc__ = self.__action["doc"]

    @property
    def name(self):
        """The name of this action."""
        return self.__action["name"]

    @property
    def fullname(self):
        """The qualified name of this action, including the handler's name."""
        return "%s.%s" % (self.__handler.name, self.name)

    @property
    def op(self):
        """The name of the underlying operation, if set."""
        return self.__action["op"]

    @property
    def is_restful(self):
        """Indicates if this action is ReSTful.

        In other words, this is a CRUD operation: create, read, update, or
        delete.
        """
        return self.__action["restful"]

    @property
    def method(self):
        """The HTTP method."""
        return self.__action["method"]

    @property
    def handler(self):
        """The `HandlerAPI`."""
        return self.__handler

    def bind(self, **params):
        """Bind URI parameters.

        :return: A `CallAPI` instance.
        """
        return CallAPI(params, self)

    @asynchronous
    async def __call__(self, **data):
        """Convenience method to do ``this.bind(**params).call(**data).data``.

        The ``params`` are extracted from the given keyword arguments.
        Whatever remains is assumed to be data to be passed to ``call()`` as
        keyword arguments.

        All keys in ``params`` that are prefixed with '_' are remapped without
        the '_' prefix into the ``call()``. This is used when the ``params``
        and data passed to the call have the same key.

        :raise KeyError: If not all required arguments are provided.

        See `CallAPI.call()` for return information and exceptions.
        """
        data = dict(data)
        params = {name: data.pop(name) for name in self.handler.params}
        for key, value in data.copy().items():
            if isinstance(value, typing.Mapping):
                del data[key]
                for nested_key, nested_value in value.items():
                    data[key + "_" + nested_key] = nested_value
        for key, value in data.copy().items():
            if key.startswith("_"):
                data[key[1:]] = data.pop(key)
        response = await self.bind(**params).call(**data)
        return response.data

    def __repr__(self):
        if self.op is None:
            return "<Action %s %s %s>" % (self.fullname, self.method, self.handler.uri)
        else:
            return "<Action %s %s %s op=%s>" % (
                self.fullname,
                self.method,
                self.handler.uri,
                self.op,
            )


CallResult = namedtuple("CallResult", ("response", "content", "data"))


class CallError(Exception):
    def __init__(self, request, response, content, call):
        desc_for_request = "%(method)s %(uri)s" % request
        desc_for_response = "HTTP %s %s" % (response.status, response.reason)
        desc_for_content = content.decode("utf-8", "replace")
        desc = "%s -> %s (%s)" % (
            desc_for_request,
            desc_for_response,
            desc_for_content
            if len(desc_for_content) <= 50
            else (desc_for_content[:49] + "…"),
        )
        super(CallError, self).__init__(desc)
        self.request = request
        self.response = response
        self.content = content
        self.call = call

    @property
    def status(self):
        return self.response.status


class CallAPI:
    def __init__(self, params, action):
        """Create a new `CallAPI`.

        :param params: Parameters to be interpolated into the action's URI.
        :param action: The `ActionAPI`.
        """
        super(CallAPI, self).__init__()
        self.__params = params
        self.__action = action
        self.__validate()

    def __validate(self):
        params_expected = self.action.handler.params
        params_observed = frozenset(self.__params)
        if params_observed != params_expected:
            if len(params_expected) == 0:
                raise TypeError("%s takes no arguments" % self.action.fullname)
            else:
                params_expected_desc = ", ".join(sorted(params_expected))
                raise TypeError(
                    "%s takes %d arguments: %s"
                    % (self.action.fullname, len(params_expected), params_expected_desc)
                )

    @property
    def action(self):
        """The `ActionAPI`."""
        return self.__action

    @property
    def uri(self):
        """The URI for this handler, with parameters interpolated."""
        # TODO: this is el-cheapo URI Template
        # <http://tools.ietf.org/html/rfc6570> support; use uritemplate-py
        # <https://github.com/uri-templates/uritemplate-py> here?
        uri = urlparse(self.action.handler.uri)
        if uri.scheme != self.action.handler.session.scheme:
            uri = uri._replace(scheme=self.action.handler.session.scheme)
        return uri.geturl().format(**self.__params)

    def rebind(self, **params):
        """Rebind the parameters into the URI.

        :return: A new `CallAPI` instance with the new parameters.
        """
        new_params = self.__params.copy()
        new_params.update(params)
        return self.__class__(new_params, self.__action)

    def call(self, **data):
        """Issue the call.

        :param data: Data to pass in the *body* of the request.
        """
        uri, body, headers = self.prepare(data)
        return self.dispatch(uri, body, headers)

    def prepare(self, data):
        """Prepare the call payload.

        This is used by `call` and can be overridden to marshal the request in
        a different way.

        :param data: Data to pass in the *body* of the request.
        :type data: dict
        """

        def expand(data):
            for name, value in data.items():
                if isinstance(value, Iterable):
                    for value in value:
                        yield name, value
                else:
                    yield name, value

        # `data` must be an iterable yielding 2-tuples.
        if self.action.method in ("GET", "DELETE"):
            # MAAS does not expect an entity-body for GET or DELETE.
            data = expand(data)
        else:
            # MAAS expects and entity-body for PUT and POST.
            data = data.items()

        # Bundle things up ready to throw over the wire.
        uri, body, headers = utils.prepare_payload(
            self.action.op, self.action.method, self.uri, data
        )

        # Headers are returned as a list, but they must be a dict for
        # the signing machinery.
        headers = dict(headers)

        # Sign request if credentials have been provided.
        credentials = self.action.handler.session.credentials
        if credentials is not None:
            utils.sign(uri, headers, credentials)

        return uri, body, headers

    @asynchronous
    async def dispatch(self, uri, body, headers):
        """Dispatch the call via HTTP.

        This is used by `call` and can be overridden to use a different HTTP
        library.
        """
        insecure = self.action.handler.session.insecure
        connector = aiohttp.TCPConnector(verify_ssl=(not insecure))
        session = aiohttp.ClientSession(connector=connector)
        async with session:
            response = await session.request(
                self.action.method, uri, data=body, headers=_prefer_json(headers)
            )
            async with response:
                # Fetch the raw body content.
                content = await response.read()

                # Debug output.
                if self.action.handler.session.debug:
                    print(response)

                # 2xx status codes are all okay.
                if response.status // 100 != 2:
                    request = {
                        "body": body,
                        "headers": headers,
                        "method": self.action.method,
                        "uri": uri,
                    }
                    raise CallError(request, response, content, self)

                # Decode from JSON if that's what it's declared as.
                if response.content_type is None:
                    data = await response.read()
                elif response.content_type.endswith("/json"):
                    data = await response.json()
                else:
                    data = await response.read()

                if response.content_type is None:
                    data = content
                elif response.content_type.endswith("/json"):
                    # JSON should always be UTF-8.
                    data = json.loads(content.decode("utf-8"))
                else:
                    data = content

                return CallResult(response, content, data)

    def __repr__(self):
        return "<Call %s @%s>" % (self.action.fullname, self.uri)


def _prefer_json(headers):
    """Prefer JSON in HTTP requests.

    If no `Accept` header has yet been defined in `headers`, this adds one
    that makes `application/json` clearly preferred. See RFC-7159 re. this
    choice of MIME type:

      The MIME media type for JSON text is application/json.

    It also happens to be what Piston understands, which is what MAAS uses.
    """
    if not any(header.lower() == "accept" for header in headers):
        headers["Accept"] = "application/json,*/*;q=0.9"
    return headers



"""Helpers to assemble and render tabular data."""

__all__ = ["Column", "RenderTarget", "Table"]

from abc import ABCMeta, abstractmethod
import collections
import csv
import enum
from io import StringIO
import json
from os import linesep

import colorclass
import terminaltables
import yaml


class RenderTarget(enum.Enum):

    # Human-readable tabluar output.
    plain = "plain"
    pretty = "pretty"

    # Machine-readable output.
    yaml = "yaml"
    json = "json"
    csv = "csv"

    def __str__(self):
        # Return the value. This makes it better to use as a choices option to
        # argparse at the command-line.
        return self.value


class Table(metaclass=ABCMeta):
    def __init__(self, *columns, visible_columns=None):
        super(Table, self).__init__()
        self.columns = collections.OrderedDict(
            (column.name, column) for column in columns
        )
        if visible_columns is None:
            self.visible_columns = collections.OrderedDict(self.columns.items())
        else:
            self.visible_columns = collections.OrderedDict(
                (column.name, column)
                for column in columns
                if column.name in visible_columns
            )

    def __getitem__(self, name):
        return self.columns[name]

    @abstractmethod
    def get_rows(self, target, data):
        """Get the rows for the table."""

    def _filter_rows(self, rows):
        """Filter `rows` based on the visible columns."""
        filtered_rows = []
        for row in rows:
            filtered_row = []
            for idx, name in enumerate(self.columns.keys()):
                if name in self.visible_columns:
                    filtered_row.append(row[idx])
            filtered_rows.append(filtered_row)
        return filtered_rows

    def render(self, target, data):
        """Render the table."""
        rows = self.get_rows(target, data)
        rows = self._filter_rows(rows)
        renderer = getattr(self, "_render_%s" % target.name, None)
        if renderer is None:
            raise ValueError("Cannot render %r for %s." % (self.value, target))
        else:
            return renderer(rows)

    def _flatten_columns(self, columns):
        cols = []
        for _, column in columns.items():
            if isinstance(column, NestedTableColumn):
                cols.extend(self._flatten_columns(column.get_columns()))
            else:
                cols.append(column)
        return cols

    def _compute_rows(self, target, data, duplicate=False):
        columns = self.visible_columns.values()
        computed = []
        for row_data in data:
            row = []
            rows = [row]
            for datum, column in zip(row_data, columns):
                if isinstance(column, NestedTableColumn):
                    nested_rows = column.get_rows(target, datum, duplicate=duplicate)
                    orig_row = list(row)
                    row.extend(nested_rows[0])
                    for nested_row in nested_rows[1:]:
                        new_row = list(orig_row)
                        if not duplicate:
                            for idx in range(len(new_row)):
                                new_row[idx] = ""
                        new_row.extend(nested_row)
                        rows.append(new_row)
                else:
                    row.append(column.render(target, datum))
            computed.extend(rows)
        return computed

    def _render_plain(self, data):
        rows = self._compute_rows(RenderTarget.plain, data)
        rows.insert(
            0, [column.title for column in self._flatten_columns(self.visible_columns)]
        )
        return terminaltables.AsciiTable(rows).table

    def _render_pretty(self, data):
        rows = self._compute_rows(RenderTarget.pretty, data)
        rows.insert(
            0, [column.title for column in self._flatten_columns(self.visible_columns)]
        )
        return terminaltables.SingleTable(rows).table

    def _render_yaml(self, data):
        columns = self.visible_columns.values()
        rows = [
            [
                column.render(RenderTarget.yaml, datum)
                for datum, column in zip(row, columns)
            ]
            for row in data
        ]
        return yaml.safe_dump(
            {
                "columns": [
                    {"name": column.name, "title": column.title} for column in columns
                ],
                "data": [
                    {column.name: datum for column, datum in zip(columns, row)}
                    for row in rows
                ],
            },
            default_flow_style=False,
        ).rstrip(linesep)

    def _render_json(self, data):
        columns = self.visible_columns.values()
        rows = [
            [
                column.render(RenderTarget.json, datum)
                for datum, column in zip(row, columns)
            ]
            for row in data
        ]
        return json.dumps(
            {
                "columns": [
                    {"name": column.name, "title": column.title} for column in columns
                ],
                "data": [
                    {column.name: datum for column, datum in zip(columns, row)}
                    for row in rows
                ],
            }
        )

    def _render_csv(self, data):
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(
            [column.name for column in self._flatten_columns(self.visible_columns)]
        )
        writer.writerows(self._compute_rows(RenderTarget.csv, data, duplicate=True))
        return output.getvalue().rstrip(linesep)

    def __repr__(self):
        return "<%s [%s]>" % (self.__class__.__name__, " ".join(self.visible_columns))


class DetailTable(Table):
    def _filter_rows(self, rows, visible_columns=None):
        """Filter `rows` based on the visible columns."""
        if visible_columns is None:
            visible_columns = self.visible_columns
        filtered_row = []
        for idx, name in enumerate(self.columns.keys()):
            if name in self.visible_columns:
                filtered_row.append(rows[idx])
        return filtered_row

    def render(self, target, data):
        renderer = getattr(self, "_render_%s" % target.name, None)
        if renderer is None:
            raise ValueError("Cannot render %r for %s." % (self.value, target))
        else:
            return renderer(data)

    def _split_nested_tables(self):
        table_columns = collections.OrderedDict(
            (name, column)
            for name, column in self.visible_columns.items()
            if isinstance(column, NestedTableColumn)
        )
        data_columns = collections.OrderedDict(
            (name, column)
            for name, column in self.visible_columns.items()
            if not isinstance(column, NestedTableColumn)
        )
        return data_columns, table_columns

    def _render_nested_table(self, target, data, column):
        data_idx = list(self.columns.keys()).index(column.name)
        data = data[data_idx]
        table = column.get_table()
        return table.render(target, data)

    def _render_nested_tables(self, target, data, columns):
        tables = [
            self._render_nested_table(target, data, column)
            for _, column in columns.items()
        ]
        if len(tables) > 0:
            return "\n" + "\n".join(tables)
        else:
            return ""

    def _render_table(self, target, terminaltable, data):
        columns, table_columns = self._split_nested_tables()
        all_rows = self.get_rows(target, data)
        rows = self._filter_rows(all_rows, visible_columns=columns)
        rows = [
            column.render(target, datum)
            for column, datum in zip(columns.values(), rows)
        ]
        table = terminaltable(
            [(column.title, datum) for column, datum in zip(columns.values(), rows)]
        )
        table.inner_heading_row_border = False
        return table.table + self._render_nested_tables(target, all_rows, table_columns)

    def _render_plain(self, data):
        return self._render_table(RenderTarget.plain, terminaltables.AsciiTable, data)

    def _render_pretty(self, data):
        return self._render_table(RenderTarget.pretty, terminaltables.SingleTable, data)

    def _render_yaml(self, data):
        columns = self.visible_columns.values()
        rows = self.get_rows(RenderTarget.yaml, data)
        rows = self._filter_rows(rows)
        rows = [
            column.render(RenderTarget.yaml, datum)
            for column, datum in zip(columns, rows)
        ]
        return yaml.safe_dump(
            {column.name: datum for column, datum in zip(columns, rows)},
            default_flow_style=False,
        ).rstrip(linesep)

    def _render_json(self, data):
        columns = self.visible_columns.values()
        rows = self.get_rows(RenderTarget.json, data)
        rows = self._filter_rows(rows)
        rows = [
            column.render(RenderTarget.json, datum)
            for column, datum in zip(columns, rows)
        ]
        return json.dumps({column.name: datum for column, datum in zip(columns, rows)})

    def _render_csv(self, data):
        columns, table_columns = self._split_nested_tables()
        all_rows = self.get_rows(RenderTarget.csv, data)
        rows = self._filter_rows(all_rows, visible_columns=columns)
        rows = [
            column.render(RenderTarget.csv, datum)
            for column, datum in zip(columns.values(), rows)
        ]
        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(
            [(column.name, datum) for column, datum in zip(columns.values(), rows)]
        )
        return output.getvalue().rstrip(linesep) + (
            self._render_nested_tables(RenderTarget.csv, all_rows, table_columns)
        )


class Column:
    def __init__(self, name, title=None):
        super(Column, self).__init__()
        self.name = name
        self.title = name if title is None else title

    def render(self, target, datum):
        if target is RenderTarget.yaml:
            return datum
        elif target is RenderTarget.json:
            return datum
        elif target is RenderTarget.csv:
            if isinstance(datum, collections.Iterable) and not isinstance(
                datum, (str, bytes)
            ):
                return ",".join(datum)
            else:
                return datum
        elif target is RenderTarget.plain:
            if datum is None:
                return ""
            elif isinstance(datum, colorclass.Color):
                return datum.value_no_colors
            elif isinstance(datum, collections.Iterable) and not isinstance(
                datum, (str, bytes)
            ):
                return "\n".join(datum)
            else:
                return str(datum)
        elif target is RenderTarget.pretty:
            if datum is None:
                return ""
            elif isinstance(datum, colorclass.Color):
                return datum
            elif isinstance(datum, collections.Iterable) and not isinstance(
                datum, (str, bytes)
            ):
                return "\n".join(datum)
            else:
                return str(datum)
        else:
            raise ValueError("Cannot render %r for %s" % (datum, target))

    def __repr__(self):
        return "<%s name=%s title=%r>" % (
            self.__class__.__name__,
            self.name,
            self.title,
        )


class NestedTableColumn(Column):
    def __init__(
        self, name, title=None, table=None, table_args=None, table_kwargs=None
    ):
        super(NestedTableColumn, self).__init__(name, title=title)
        self.table = table
        self.table_args = table_args
        self.table_kwargs = table_kwargs
        if table is None:
            raise ValueError("table is required.")

    def get_table(self):
        table_args = self.table_args
        if table_args is None:
            table_args = []
        table_kwargs = self.table_kwargs
        if table_kwargs is None:
            table_kwargs = {}
        return self.table(*table_args, **table_kwargs)

    def get_columns(self):
        return self.get_table().visible_columns

    def get_rows(self, target, data, duplicate=False):
        table = self.get_table()
        rows = table.get_rows(target, data)
        rows = table._filter_rows(rows)
        rows = table._compute_rows(target, rows, duplicate=duplicate)
        if len(rows) == 0:
            # Nested table column must always return one row even if its
            # an empty row.
            rows = [[" " for _ in range(len(table.visible_columns))]]
        return rows

    def render(self, target, datum):
        table = self.get_table()
        if target is RenderTarget.yaml:
            return yaml.safe_load(table.render(target, datum))
        elif target is RenderTarget.json:
            return json.loads(table.render(target, datum))
        else:
            raise ValueError(
                "Should not be called on a nested table column, the "
                "table render should handle this correctly."
            )



"""Objects for RAIDs."""

__all__ = ["Raid", "Raids"]

from typing import Iterable, Union

from . import check, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType, to
from .nodes import Node
from .block_devices import BlockDevice
from .partitions import Partition
from .filesystem_groups import DevicesField, FilesystemGroup
from ..enum import RaidLevel


class RaidType(ObjectType):
    """Metaclass for `Raid`."""

    async def read(cls, node, id):
        """Get `Raid` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


class Raid(FilesystemGroup, metaclass=RaidType):
    """A RAID on a machine."""

    uuid = ObjectField.Checked("uuid", check(str), check(str))

    level = ObjectField.Checked("level", to(RaidLevel), readonly=True)
    size = ObjectField.Checked("size", check(int), check(int), readonly=True)

    devices = DevicesField("devices")
    spare_devices = DevicesField("spare_devices")
    virtual_device = ObjectFieldRelated(
        "virtual_device", "BlockDevice", reverse=None, readonly=True
    )

    def __repr__(self):
        return super(Raid, self).__repr__(fields={"name", "level", "size"})

    async def delete(self):
        """Delete this RAID."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)


class RaidsType(ObjectType):
    """Metaclass for `Raids`."""

    async def read(cls, node):
        """Get list of `Raid`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls,
        node: Union[Node, str],
        level: Union[RaidLevel, str],
        devices: Iterable[Union[BlockDevice, Partition]],
        *,
        name: str = None,
        uuid: str = None,
        spare_devices: Iterable[Union[BlockDevice, Partition]]
    ):
        """
        Create a RAID on a Node.

        :param node: Node to create the interface on.
        :type node: `Node` or `str`
        :param level: RAID level.
        :type level: `RaidLevel`
        :param devices: Mixed list of block devices or partitions to create
            the RAID from.
        :type devices: iterable of mixed type of `BlockDevice` or `Partition`
        :param name: Name of the RAID (optional).
        :type name: `str`
        :param uuid: The UUID for the RAID (optional).
        :type uuid: `str`
        :param spare_devices: Mixed list of block devices or partitions to add
            as spare devices on the RAID.
        :type spare_devices: iterable of mixed type of `BlockDevice` or
            `Partition`
        """
        if isinstance(level, RaidLevel):
            level = level.value
        params = {"level": str(level)}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )

        if len(devices) == 0:
            raise ValueError("devices must contain at least one device.")

        block_devices = []
        partitions = []
        for idx, device in enumerate(devices):
            if isinstance(device, BlockDevice):
                block_devices.append(device.id)
            elif isinstance(device, Partition):
                partitions.append(device.id)
            else:
                raise TypeError(
                    "devices[%d] must be a BlockDevice or "
                    "Partition, not %s" % type(device).__name__
                )
        if len(block_devices) > 0:
            params["block_devices"] = block_devices
        if len(partitions) > 0:
            params["partitions"] = partitions

        spare_block_devices = []
        spare_partitions = []
        for idx, device in enumerate(spare_devices):
            if isinstance(device, BlockDevice):
                spare_block_devices.append(device.id)
            elif isinstance(device, Partition):
                spare_partitions.append(device.id)
            else:
                raise TypeError(
                    "spare_devices[%d] must be a BlockDevice or "
                    "Partition, not %s" % type(device).__name__
                )
        if len(spare_block_devices) > 0:
            params["spare_devices"] = spare_block_devices
        if len(spare_partitions) > 0:
            params["spare_partitions"] = spare_partitions

        if name is not None:
            params["name"] = name
        if uuid is not None:
            params["uuid"] = uuid
        return cls._object(await cls._handler.create(**params))


class Raids(ObjectSet, metaclass=RaidsType):
    """The set of RAIDs on a machine."""

    @property
    def by_name(self):
        """Return mapping of name to `Raid`."""
        return {raid.name: raid for raid in self}

    def get_by_name(self, name):
        """Return a `Raid` by its name."""
        return self.by_name[name]



"""Commands for spaces."""

__all__ = ["register"]

from http import HTTPStatus

from . import CommandError, OriginPagedTableCommand, tables
from ..bones import CallError
from ..utils.maas_async import asynchronous


class cmd_spaces(OriginPagedTableCommand):
    """List spaces."""

    def __init__(self, parser):
        super(cmd_spaces, self).__init__(parser)
        parser.add_argument(
            "--minimal", action="store_true", help=("Output only the space names.")
        )

    @asynchronous
    async def load_object_sets(self, origin):
        spaces = origin.Spaces.read()
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await spaces, await fabrics, await subnets

    def execute(self, origin, options, target):
        visible_columns = None
        if options.minimal:
            visible_columns = ("name",)
        spaces, fabrics, subnets = self.load_object_sets(origin)
        table = tables.SpacesTable(
            visible_columns=visible_columns, fabrics=fabrics, subnets=subnets
        )
        return table.render(target, spaces)


class cmd_space(OriginPagedTableCommand):
    """Details of a space."""

    def __init__(self, parser):
        super(cmd_space, self).__init__(parser)
        parser.add_argument("name", nargs=1, help=("Name of the space."))

    @asynchronous
    async def load_object_sets(self, origin):
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await fabrics, await subnets

    def execute(self, origin, options, target):
        try:
            space = origin.Space.read(options.name[0])
        except CallError as error:
            if error.status == HTTPStatus.NOT_FOUND:
                raise CommandError("Unable to find space %s." % options.name[0])
            else:
                raise
        fabrics, subnets = self.load_object_sets(origin)
        table = tables.SpaceDetail(fabrics=fabrics, subnets=subnets)
        return table.render(target, space)


def register(parser):
    """Register commands with the given parser."""
    cmd_spaces.register(parser)
    cmd_space.register(parser)



"""Objects for domains."""

__all__ = ["Domain", "Domains"]

from . import check, check_optional, Object, ObjectField, ObjectSet, ObjectType


class DomainType(ObjectType):
    """Metaclass for `Domains`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(cls, name: str, authoritative: bool = True, ttl: int = None):
        """
        Create a `Domain` in MAAS.

        :param name: The name of the `Domain`.
        :type name: `str`
        :param authoritative: Whether the domain is authoritative.
        :type authoritative: `bool`
        :param ttl: Optional TTL for the domain.
        :type ttl: `int`
        :returns: The created `Domain`
        :rtype: `Domain`
        """
        params = {"name": name, "authoritative": authoritative}
        if ttl is not None:
            params["ttl"] = ttl
        return cls._object(await cls._handler.create(**params))


class Domains(ObjectSet, metaclass=DomainType):
    """The set of domains stored in MAAS."""


class DomainType(ObjectType):
    async def read(cls, id):
        data = await cls._handler.read(id=id)
        return cls(data)


class Domain(Object, metaclass=DomainType):
    """A domain stored in MAAS."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    name = ObjectField.Checked("name", check(str), check(str))
    authoritative = ObjectField.Checked(
        "authoritative", check_optional(bool), check_optional(bool)
    )
    ttl = ObjectField.Checked("ttl", check_optional(int), check_optional(int))

    def __repr__(self):
        return super(Domain, self).__repr__(fields={"name", "authoritative", "ttl"})

    async def delete(self):
        """
        Deletes the `domain` from MAAS.
        """
        return await self._handler.delete(id=self.id)



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Profile configuration."""

__all__ = ["Profile", "ProfileStore", "ProfileNotFound"]

from contextlib import contextmanager
from copy import deepcopy
import json
from pathlib import Path
import sqlite3
from textwrap import dedent
import typing

from . import api_url
from .creds import Credentials
from .types import JSONObject


class Profile(tuple):
    """A profile is all that's required to talk to a remote MAAS."""

    __slots__ = ()

    def __new__(
        cls,
        name: str,
        url: str,
        *,
        credentials: typing.Union[Credentials, typing.Sequence, str, None],
        description: dict,
        **other: JSONObject
    ):
        return super(Profile, cls).__new__(
            cls,
            (name, api_url(url), Credentials.parse(credentials), description, other),
        )

    @property
    def name(self) -> str:
        """The name of this profile."""
        return self[0]

    @property
    def url(self) -> str:
        """The URL for this profile."""
        return self[1]

    @property
    def credentials(self) -> typing.Optional[Credentials]:
        """The credentials for this profile, if set."""
        return self[2]

    @property
    def description(self) -> dict:
        """The description for this profile, detailing remote resources et al.

        A deep copy is made of this before it is returned, so that mutations
        will not alter the originating copy.
        """
        return deepcopy(self[3])

    @property
    def other(self) -> dict:
        """Other fields that have been stored for this profile.

        A deep copy is made of this before it is returned, so that mutations
        will not alter the originating copy.
        """
        return deepcopy(self[-1])

    def replace(self, **updates):
        """Return a new profile with the given updates.

        Unspecified fields will be the same as this instance. See `__new__`
        for details on the arguments.
        """
        state = self.dump()
        state.update(updates)
        return self.__class__(**state)

    def dump(self):
        """Return a dict of fields that can be used to recreate this profile.

        For example::

          >>> profile = Profile(name="foobar", ...)
          >>> profile == Profile(**profile.dump())
          True

        Use this value when persisting a profile.
        """
        return dict(
            self.other,
            name=self.name,
            url=self.url,
            credentials=self.credentials,
            description=self.description,
        )

    def __repr__(self):
        if self.credentials is None:
            return "<%s %s (anonymous) %s>" % (
                self.__class__.__name__,
                self.name,
                self.url,
            )
        else:
            return "<%s %s %s>" % (self.__class__.__name__, self.name, self.url)


class ProfileNotFound(Exception):
    """The named profile was not found."""

    def __init__(self, name):
        super(ProfileNotFound, self).__init__("Profile '%s' not found." % (name,))


def schema_create(conn):
    """Create the index for storing profiles.

    This is idempotent; it can be called every time a database is opened to
    make sure it's ready to use and up-to-date.

    :param conn: A connection to an SQLite3 database.
    """
    conn.execute(
        dedent(
            """\
    CREATE TABLE IF NOT EXISTS profiles
      (id INTEGER PRIMARY KEY,
       name TEXT NOT NULL UNIQUE,
       data BLOB NOT NULL,
       selected BOOLEAN NOT NULL DEFAULT FALSE)
    """
        )
    )
    # Partial indexes are only available in >=3.8.0 and expressions in indexes
    # are only available in >=3.9.0 (https://www.sqlite.org/partialindex.html
    # & https://www.sqlite.org/expridx.html). Don't bother with any kind of
    # index before that because it would complicate upgrades.
    if sqlite3.sqlite_version_info >= (3, 9, 0):
        # This index is for data integrity -- ensuring that only one profile
        # is the default ("selected") profile -- and speed a distant second.
        conn.execute(
            dedent(
                """\
        CREATE UNIQUE INDEX IF NOT EXISTS
          only_one_profile_selected ON profiles
          (selected IS NOT NULL) WHERE selected
        """
            )
        )


def schema_import(conn, dbpath):
    """Import profiles from another database.

    This does not overwrite existing profiles in the target database. Profiles
    in the source database that share names with those in the target database
    are ignored.

    :param conn: A connection to an SQLite3 database into which to copy
        profiles.
    :param dbpath: The filesystem path to the source SQLite3 database.
    """
    conn.execute("ATTACH DATABASE ? AS source", (str(dbpath),))
    conn.execute(
        "INSERT OR IGNORE INTO profiles (name, data)"
        " SELECT name, data FROM source.profiles"
        " WHERE data IS NOT NULL"
    )
    conn.commit()  # need to commit before detaching the other db
    conn.execute("DETACH DATABASE source")


class ProfileStore:
    """Store profile configurations in an sqlite3 database."""

    def __init__(self, database):
        self.database = database
        schema_create(database)

    def __iter__(self):
        results = self.database.execute("SELECT name FROM profiles").fetchall()
        return (name for (name,) in results)

    def load(self, name: str) -> Profile:
        found = self.database.execute(
            "SELECT data FROM profiles" " WHERE name = ?", (name,)
        ).fetchone()
        if found is None:
            raise ProfileNotFound(name)
        else:
            state = json.loads(found[0])
            state["name"] = name  # Belt-n-braces.
            return Profile(**state)

    def save(self, profile: Profile):
        state = profile.dump()
        data = json.dumps(state)
        # On the face of it `INSERT OR REPLACE` would be an obvious way to do
        # this. However, on conflict it will erase the value of the `selected`
        # column to the default (which is false). Hence we do it in two steps,
        # within a transaction.
        with self.database:
            # Ensure there's a row for this profile.
            self.database.execute(
                "INSERT OR IGNORE INTO profiles (name, data) VALUES (?, '')",
                (profile.name,),
            )
            # Update the row's data.
            self.database.execute(
                "UPDATE profiles SET data = ? WHERE name = ?", (data, profile.name)
            )

    def delete(self, name: str):
        self.database.execute("DELETE FROM profiles WHERE name = ?", (name,))

    @property
    def default(self) -> typing.Optional[Profile]:
        """The name of the default profile to use, or `None`."""
        found = self.database.execute(
            "SELECT name, data FROM profiles WHERE selected" " ORDER BY name LIMIT 1"
        ).fetchone()
        if found is None:
            return None
        else:
            state = json.loads(found[1])
            state["name"] = found[0]  # Belt-n-braces.
            return Profile(**state)

    @default.setter
    def default(self, profile: Profile):
        with self.database:
            self.save(profile)
            del self.default
            self.database.execute(
                "UPDATE profiles SET selected = (name = ?)", (profile.name,)
            )

    @default.deleter
    def default(self):
        self.database.execute("UPDATE profiles SET selected = 0")

    @classmethod
    @contextmanager
    def open(
        cls,
        dbpath=Path("~/.maas.db").expanduser(),
        migrate_from=Path("~/.maascli.db").expanduser(),
    ):
        """Load a profiles database.

        Called without arguments this will open (and create) a database in the
        user's home directory.

        **Note** that this returns a context manager which will close the
        database on exit, saving if the exit is clean.

        :param dbpath: The path to the database file to create and open.
        :param migrate_from: Path to a database file to migrate from.
        """
        # Ensure we're working with a Path instance.
        dbpath = Path(dbpath)
        migrate_from = Path(migrate_from)
        # See if we ought to do a one-time migration.
        migrate = migrate_from.is_file() and not dbpath.exists()
        # Initialise filename with restrictive permissions...
        dbpath.touch(mode=0o600, exist_ok=True)
        # Final check to see if it's safe to migrate.
        migrate = migrate and not migrate_from.samefile(dbpath)
        # before opening it with sqlite.
        database = sqlite3.connect(str(dbpath))
        try:
            store = cls(database)
            if migrate:
                schema_import(database, migrate_from)
                yield store
            else:
                yield store
        except:  # noqa: E722
            raise
        else:
            database.commit()
        finally:
            database.close()



"""Interact with a remote MAAS (https://maas.ubuntu.com/).

These are highish-level bindings that provide user-oriented objects to
interact with. They're intended for developers, for folk comfortable working
with Python. They're not a front-end. Hence the name "viscera"; the classes
and functions here are the meat and organs around the "bones", but not the
"flesh".

"""

__all__ = [
    "check",
    "check_optional",
    "Object",
    "ObjectField",
    "ObjectMethod",
    "ObjectType",
    "Origin",
    "OriginBase",
]

from collections.abc import Iterable, Mapping, Sequence
from collections import defaultdict
from copy import copy
from datetime import datetime
from functools import wraps
from importlib import import_module
from itertools import chain, starmap
from types import MethodType

import pytz

from .. import bones
from ..errors import ObjectNotLoaded
from ..utils import get_all_subclasses, vars_class
from ..utils.maas_async import Asynchronous


undefined = object()


class Disabled:
    """A disabled method."""

    def __init__(self, name, alternative=None):
        super(Disabled, self).__init__()
        self.name, self.alternative = name, alternative

    def __call__(self, *args, **kwargs):
        if self.alternative is None:
            raise RuntimeError("%s has been disabled" % (self.name,))
        else:
            raise RuntimeError(
                "%s has been disabled; use %s instead" % (self.name, self.alternative)
            )


def dir_class(cls):
    """Return a list of names available on `cls`.

    Eliminates names that bind to an `ObjectMethod` without a corresponding
    class method; see `ObjectMethod`.
    """
    # Class attributes (including methods).
    for name, value in vars_class(cls).items():
        if isinstance(value, ObjectMethod):
            if value.has_classmethod:
                yield name
        elif isinstance(value, Disabled):
            pass  # Hide this; disabled.
        else:
            yield name
    # Metaclass attributes.
    for name, value in vars_class(type(cls)).items():
        if name == "mro":
            pass  # Hide this; not interesting.
        elif isinstance(value, Disabled):
            pass  # Hide this; disabled.
        else:
            yield name


def dir_instance(inst):
    """Return a list of names available on `inst`.

    Eliminates names that bind to an `ObjectMethod` without a corresponding
    instance method; see `ObjectMethod`.
    """
    # Skip instance attributes; __slots__ is automatically defined, and
    # descriptors are used to define attributes. Instead, go straight to class
    # attributes (including methods).
    for name, value in vars_class(type(inst)).items():
        if isinstance(value, ObjectMethod):
            if value.has_instancemethod:
                yield name
        elif isinstance(value, Disabled):
            pass  # Hide this; disabled.
        elif isinstance(value, (classmethod, staticmethod)):
            pass  # Hide this; not interesting here.
        else:
            yield name


class OriginObjectRef:
    """A reference to an object in the origin.

    Use this on an `Object` to reference a related object type via a bound
    origin (see `OriginBase`). By default this guesses that the name of the
    referenced object is the singular of the owner's name. For example,
    given::

      class Things(Object):
          _object = OriginObjectRef()

    `Things._object` would look for `Things._origin.Thing`. When referencing
    something that has non-trivial singular/plural naming, specify the name to
    the `OriginObjectRef` constructor.
    """

    def __init__(self, name=None):
        super(OriginObjectRef, self).__init__()
        self.name = name

    def __get__(self, instance, owner):
        if self.name is None:
            name = owner.__name__.split(".")[0]
            return getattr(owner._origin, name.rstrip("s"))
        else:
            return getattr(owner._origin, self.name)

    def __set__(self, instance, value):
        raise AttributeError()


class ObjectType(Asynchronous, metaclass=Asynchronous):
    def __dir__(cls):
        return dir_class(cls)

    def __new__(cls, name, bases, attrs):
        attrs.setdefault("__slots__", ())
        return super(ObjectType, cls).__new__(cls, name, bases, attrs)

    def bind(cls, origin, handler, handlers, *, name=None):
        """Bind this object to the given origin and handler.

        :param origin: An instance of `Origin`.
        :param handler: An instance of `bones.HandlerAPI`.
        :param handlers: All handlers from `bones`.
        :return: A subclass of this class.
        """
        name = cls.__name__ if name is None else name
        attrs = {
            "_origin": origin,
            "_handler": handler,
            "__module__": "origin",  # Could do better?
        }
        return type(name, (cls,), attrs)


class ObjectBasics:

    __slots__ = ()

    def __dir__(self):
        return dir_instance(self)

    def __str__(self):
        return self.__class__.__qualname__


def is_pk_descriptor(descriptor, include_alt=False):
    """Return true if `descriptor` is a primary key."""
    if descriptor.pk is True or type(descriptor.pk) is int:
        return True
    if include_alt:
        return descriptor.alt_pk is True or type(descriptor.alt_pk) is int
    else:
        return False


def get_pk_descriptors(cls):
    """Return tuple of tuples with attribute name and descriptor on the
    `cls` that is defined as the primary keys."""
    pk_fields = {
        name: descriptor
        for name, descriptor in vars_class(cls).items()
        if isinstance(descriptor, ObjectField) and is_pk_descriptor(descriptor)
    }
    alt_pk_fields = defaultdict(list)
    for name, descriptor in vars_class(cls).items():
        if isinstance(descriptor, ObjectField):
            if descriptor.alt_pk is True:
                alt_pk_fields[0].append((name, descriptor))
            elif type(descriptor.alt_pk) is int:
                alt_pk_fields[descriptor.alt_pk].append((name, descriptor))
    if len(pk_fields) == 1:
        return ((pk_fields.popitem(),), (alt_pk_fields[0],))
    elif len(pk_fields) > 1:
        unique_pk_fields = {
            name: descriptor
            for name, descriptor in pk_fields.items()
            if descriptor.pk is True
        }
        if unique_pk_fields:
            raise AttributeError(
                "more than one field is marked as unique primary key: %s"
                % (", ".join(sorted(pk_fields)))
            )
        pk_descriptors = tuple(
            sorted(
                ((name, descriptor) for name, descriptor in pk_fields.items()),
                key=lambda item: item[1].pk,
            )
        )
        alt_pk_descriptors = tuple(
            alt_pk_fields[idx] for idx, (name, descriptor) in enumerate(pk_descriptors)
        )
        return pk_descriptors, alt_pk_descriptors
    else:
        return tuple(), tuple()


def set_alt_pk_value(alt_descriptors, obj_data, data):
    for name, descriptor in alt_descriptors:
        if descriptor.name in data:
            obj_data[descriptor.name] = data[descriptor.name]
            return name
    return None


class Object(ObjectBasics, metaclass=ObjectType):
    """An object in a MAAS installation."""

    __slots__ = ("__weakref__", "_data", "_orig_data", "_changed_data", "_loaded")

    def __init__(self, data, local_data=None):
        super(Object, self).__init__()
        self._data = data
        self._changed_data = {}
        self._loaded = False
        if isinstance(data, Mapping) and not data.get("__incomplete__", False):
            self._reset(data)
            self._loaded = True
        else:
            descriptors, alt_descriptors = get_pk_descriptors(type(self))
            if len(descriptors) == 1:
                if isinstance(data, Mapping):
                    new_data = {}
                    try:
                        new_data[descriptors[0][1].name] = data[descriptors[0][1].name]
                    except KeyError:
                        found_name = set_alt_pk_value(
                            alt_descriptors[0], new_data, data
                        )
                        if found_name:
                            # Validate that the set data is correct and
                            # can be converted to the python value.
                            getattr(self, found_name)
                        else:
                            raise
                    else:
                        # Validate that the set data is correct and
                        # can be converted to the python value.
                        getattr(self, descriptors[0][0])
                    self._reset(new_data)
                else:
                    self._reset({descriptors[0][1].name: data})
                    # Validate that the primary key is correct and
                    # can be converted to the python value.
                    getattr(self, descriptors[0][0])
            elif len(descriptors) > 1:
                if isinstance(data, Mapping):
                    obj_data = {}
                    found_names = []
                    for idx, (name, descriptor) in enumerate(descriptors):
                        try:
                            obj_data[descriptor.name] = data[descriptor.name]
                        except KeyError:
                            found_name = set_alt_pk_value(
                                alt_descriptors[idx], obj_data, data
                            )
                            if found_name:
                                found_names.append(found_name)
                            else:
                                raise
                        else:
                            found_names.append(name)
                    self._reset(obj_data)
                    # Validate that all set data is correct and can be
                    # converted to the python value.
                    for name in found_names:
                        getattr(self, name)
                elif isinstance(data, Sequence):
                    obj_data = {}
                    for idx, (name, descriptor) in enumerate(descriptors):
                        obj_data[descriptor.name] = data[idx]
                    self._reset(obj_data)
                    for name, _ in descriptors:
                        # Validate that the primary key is correct and
                        # can be converted to the python value.
                        getattr(self, name)
                else:
                    raise TypeError(
                        "data must be a mapping or a sequence, not %s"
                        % (type(data).__name__)
                    )
            else:
                if not isinstance(data, Mapping):
                    raise TypeError(
                        "data must be a mapping, not %s" % type(data).__name__
                    )
                else:
                    raise ValueError(
                        "data cannot be incomplete without any primary keys " "defined"
                    )
        if local_data is not None:
            if isinstance(local_data, Mapping):
                self._data.update(local_data)
            else:
                raise TypeError(
                    "local_data must be a mapping, not %s" % type(data).__name__
                )

    def _reset(self, data):
        """Reset the object to look pristine with the given data.

        All tracked changes will be discarded, and the object will be
        ready to track new changes.
        """
        self._data = data
        # Make a shallow copy of each item in the original data so that
        # we can keep track of the changes. A shallow copy is enough,
        # since we only care about changes that are directly related to
        # this object.
        self._orig_data = {key: copy(value) for key, value in data.items()}
        self._changed_data = {}

    def __getattribute__(self, name):
        """Prevent access to fields that are not defined as primary keys on
        the object when its not loaded."""
        fields = {
            name: descriptor
            for name, descriptor in vars_class(type(self)).items()
            if isinstance(descriptor, ObjectField)
        }
        if name in fields:
            if super().__getattribute__("_loaded"):
                return super(Object, self).__getattribute__(name)
            elif is_pk_descriptor(fields[name], include_alt=True):
                return super(Object, self).__getattribute__(name)
            else:
                raise ObjectNotLoaded(
                    "cannot access attribute '%s' of object '%s'"
                    % (name, type(self).__name__)
                )
        else:
            return super(Object, self).__getattribute__(name)

    def __eq__(self, other):
        """Strict equality check.

        The type of `other` must exactly match the type of `self`. All data
        must also match.
        """
        return type(self) is type(other) and self._data == other._data

    def __repr__(self, *, name=None, fields=None):
        if name is None:
            name = self.__class__.__name__
        unloaded = ""
        if not self.loaded:
            unloaded = " (unloaded)"
            descriptors, alt_descriptors = get_pk_descriptors(type(self))
            if descriptors:
                fields = []
                for idx, (field_name, _) in enumerate(descriptors):
                    if hasattr(self, field_name):
                        fields.append(field_name)
                    else:
                        for alt_field_name, _ in alt_descriptors[idx]:
                            if hasattr(self, alt_field_name):
                                fields.append(alt_field_name)
            else:
                fields = []
        if fields is None:
            fields = sorted(
                name
                for name, value in vars_class(type(self)).items()
                if isinstance(value, ObjectField)
            )
        else:
            fields = sorted(fields)
        values = (getattr(self, name) for name in fields)
        pairs = starmap("{0}={1!r}".format, zip(fields, values))
        desc = " ".join(pairs)
        if len(desc) == 0:
            return "<%s%s>" % (name, unloaded)
        else:
            return "<%s %s%s>" % (name, desc, unloaded)

    def __hash__(self):
        name = str(self.__class__.__name__)
        if hasattr(self, "id"):
            return hash((name, self.id))
        if hasattr(self, "system_id"):
            return hash((name, self.system_id))
        return None

    @property
    def loaded(self):
        """True when the object is loaded.

        Accessing any attribute (expected for the primary keys) of an unloaded
        object will raise an `ObjectNotLoaded` exception.
        """
        return self._loaded

    async def refresh(self):
        """Refresh the object from MAAS."""
        cls = type(self)
        if hasattr(cls, "read"):
            descriptors, alt_descriptors = get_pk_descriptors(cls)
            if len(descriptors) == 1:
                try:
                    obj = await cls.read(getattr(self, descriptors[0][0]))
                except AttributeError:
                    found = False
                    for alt_name, _ in alt_descriptors[0]:
                        if hasattr(self, alt_name):
                            obj = await cls.read(getattr(self, alt_name))
                            found = True
                            break
                    if not found:
                        raise
            elif len(descriptors) > 1:
                args = []
                for idx, (name, _) in enumerate(descriptors):
                    try:
                        args.append(getattr(self, name))
                    except AttributeError:
                        found = False
                        for alt_name, _ in alt_descriptors[idx]:
                            if hasattr(self, alt_name):
                                args.append(getattr(self, alt_name))
                                found = True
                                break
                        if not found:
                            raise
                obj = await cls.read(*args)
            else:
                raise AttributeError(
                    "unable to perform 'refresh' no primary key " "fields defined."
                )
            if type(obj) is cls:
                self._reset(obj._data)
                self._loaded = True
            else:
                raise TypeError(
                    "result of '%s.read' must be '%s', not '%s'"
                    % (cls.__name__, cls.__name__, type(obj).__name__)
                )
        else:
            raise AttributeError("'%s' object doesn't support refresh." % cls.__name__)

    async def save(self):
        """Save the object in MAAS."""
        if hasattr(self._handler, "update"):
            if self._changed_data:
                update_data = dict(self._changed_data)
                update_data.update(
                    {key: self._orig_data[key] for key in self._handler.params}
                )
                data = await self._handler.update(**update_data)
                self._reset(data)
        else:
            raise AttributeError(
                "'%s' object doesn't support save." % type(self).__name__
            )


def ManagedCreate(super_cls):
    """Dynamically creates a `create` method for a `ObjectSet.Managed` class
    that calls the `super_cls.create`.

    The first positional argument that is passed to the `super_cls.create` is
    the `_manager` that was set using `ObjectSet.Managed`. The created object
    is added to the `ObjectSet.Managed` also placed in the correct
    `_data[field]` and `_orig_data[field]` for the `_manager` object.
    """

    @wraps(super_cls.create)
    async def _create(self, *args, **kwargs):
        cls = type(self)
        manager = getattr(cls, "_manager", None)
        manager_field = getattr(cls, "_manager_field", None)
        if manager is not None and manager_field is not None:
            args = (manager,) + args
            new_obj = await super_cls.create(*args, **kwargs)
            self._items = self._items + [new_obj]
            manager._data[manager_field.name] = manager._data[manager_field.name] + [
                new_obj._data
            ]
            manager._orig_data[manager_field.name] = manager._orig_data[
                manager_field.name
            ] + [new_obj._data]
            return new_obj
        else:
            raise AttributeError(
                "create is not supported; %s is not a managed set"
                % (super_cls.__name__)
            )

    return _create


class ObjectSet(ObjectBasics, metaclass=ObjectType):
    """A set of objects in a MAAS installation."""

    __slots__ = "__weakref__", "_items"

    _object = OriginObjectRef()

    @classmethod
    def Managed(cls, manager, field, items):
        """Create a custom `ObjectSet` that is managed by a related `Object.`

        :param manager: The manager of the `ObjectSet`. This is the `Object`
            that manages this set of objects.
        :param field: The field on the `manager` that created this managed
            `ObjectSet`.
        :param items: The items in the `ObjectSet`.
        """
        attrs = {"_manager": manager, "_manager_field": field}
        if hasattr(cls, "create"):
            attrs["create"] = ManagedCreate(cls)
        cls = type(
            "%s.Managed#%s" % (cls.__name__, manager.__class__.__name__), (cls,), attrs
        )
        return cls(items)

    def __init__(self, items):
        super(ObjectSet, self).__init__()
        if isinstance(items, (Mapping, str, bytes)):
            raise TypeError("data must be sequence-like, not %s" % type(items).__name__)
        elif isinstance(items, Sequence):
            self._items = items
        elif isinstance(items, Iterable):
            self._items = list(items)
        else:
            raise TypeError("data must be sequence-like, not %s" % type(items).__name__)

    def __len__(self):
        """Return the count of items contained herein."""
        return len(self._items)

    def __getitem__(self, spec):
        """Get a contained item or slice of contained items.

        Fetching a slice returns a new instance of this class containing the
        subset of items defined by the slice.
        """
        if isinstance(spec, slice):
            self = copy(self)
            self._items = self._items[spec]
            return self
        else:
            return self._items[spec]

    def __iter__(self):
        """Return an iterator for the contained items."""
        return iter(self._items)

    def __reversed__(self):
        """Efficiently provide the contained items in reversed order.

        This is more efficient than relying on the default behaviour of
        ``reversed``, which is to use ``__len__`` and ``__getitem__``.
        """
        return reversed(self._items)

    def __contains__(self, item):
        """Efficiently test if a given item is among the contained items.

        This is more efficient than relying on the default behaviour of
        ``in``, which is to use ``__iter__`` then the old sequence iteration
        protocol using ``__len__`` and ``__getitem__``.
        """
        return item in self._items

    def __eq__(self, other):
        """Strict equality check.

        The type of `other` must exactly match the type of `self`. All items
        must also match.
        """
        return type(self) is type(other) and self._items == other._items

    def __repr__(self):
        return "<%s length=%d items=%r>" % (
            self.__class__.__name__,
            len(self._items),
            self._items,
        )


class ObjectField:
    """A field on an `Object`.

    By default, no conversion and/or validation is performed; the value
    retrieved from MAAS's Web API (tranmitted as JSON) is returned when
    accessing this field, and is set or deleted when setting or deleting this
    field.

    It is possible to declare this field as read-only and to specify a default
    value when the corresponding datum is not found.

    A word on the "value" and "datum" nomenclature used here:

    * A "value" is a Python-side value, i.e. the one you'll work with in a
      program that uses this API.

    * A "datum" is an item of data obtained from MAAS's Web API, or ready to
      be sent back to MAAS's Web API. It is almost always something that can
      be dumped as JSON.

    A value passed into a field must be converted to a datum and vice-versa.
    To support this, two methods can be customised in `ObjectField`
    subclasses: ``value_to_datum`` and ``datum_to_value``.

    These methods serve to both convert and validate, but the default
    implementions do nothing.

    Alternatively, simple conversion functions can be passed to the `Checked`
    constructor.

    """

    @classmethod
    def Checked(cls, name, datum_to_value=None, value_to_datum=None, **other):
        """Create a custom `ObjectField` that validates values and datums.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        :param datum_to_value: A callable taking a single ``datum`` argument,
            passed positionally. This callable should convert the datum to a
            Python-side value, and/or raise an exception for invalid datums.
        :param value_to_datum: A callable taking a single ``value`` argument,
            passed positionally. This callable should convert the value to a
            MAAS-side datum, and/or raise an exception for invalid values.
        :param other: Additional arguments to pass to the default
            `ObjectField` constructor.
        """
        attrs = {}
        if datum_to_value is not None:

            @wraps(datum_to_value)
            def datum_to_value_method(instance, datum):
                return datum_to_value(datum)

            attrs["datum_to_value"] = staticmethod(datum_to_value_method)
        if value_to_datum is not None:

            @wraps(value_to_datum)
            def value_to_datum_method(instance, value):
                return value_to_datum(value)

            attrs["value_to_datum"] = staticmethod(value_to_datum_method)
        cls = type("%s.Checked#%s" % (cls.__name__, name), (cls,), attrs)
        return cls(name, **other)

    def __init__(
        self, name, *, default=undefined, readonly=False, pk=False, alt_pk=False
    ):
        """Create a `ObjectField` with an optional default.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        :param default: A default value to return when `name` is not found in
            the MAAS-side data dictionary.
        :param readonly: If true, prevent setting or deleting of this field.
        :param pk: If true marks the field as the unique primary key for the
            object it is defined on. If an integer then it define its place
            in the tuple of values that makes the object uniquely identified.
        :param alt_pk: If true marks the field as an alternative unique
            primary key for this object it is defined on. If an integer then
            it define its place in the tuple of values that makes the object
            uniquely identified.
        """
        super(ObjectField, self).__init__()
        self.name = name
        self.default = default
        if not isinstance(readonly, bool):
            raise TypeError("readonly must be a bool, not %s" % type(readonly).__name__)
        else:
            self.readonly = readonly
        if not isinstance(pk, (bool, int)):
            raise TypeError("pk must be a bool or an int, not %s" % type(pk).__name__)
        else:
            self.pk = pk
        if self.pk is not False and alt_pk is not False:
            raise ValueError("pk and alt_pk cannot be defined on the same field.")
        elif not isinstance(alt_pk, (bool, int)):
            raise TypeError(
                "alt_pk must be a bool or an int, not %s" % (type(alt_pk).__name__)
            )
        else:
            self.alt_pk = alt_pk

    def datum_to_value(self, instance, datum):
        """Convert a given MAAS-side datum to a Python-side value.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The MAAS-side datum to validate and convert into a
            Python-side value.
        :return: A value derived from the given datum.
        """
        return datum

    def value_to_datum(self, instance, value):
        """Convert a given Python-side value to a MAAS-side datum.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The Python-side value to validate and convert into a
            MAAS-side datum.
        :return: A datum derived from the given value.
        """
        return value

    def __get__(self, instance, owner):
        """Standard Python descriptor method."""
        if instance is None:
            return self
        else:
            if self.name in instance._data:
                datum = instance._data[self.name]
                return self.datum_to_value(instance, datum)
            elif self.default is undefined:
                raise AttributeError(self.name)
            else:
                return self.default

    def __set__(self, instance, value):
        """Standard Python descriptor method."""
        if self.readonly:
            raise AttributeError("%s is read-only" % self.name)
        else:
            datum = self.value_to_datum(instance, value)
            instance._data[self.name] = datum
            if self.name in instance._orig_data:
                orig_datum = instance._orig_data[self.name]
                if self.name in instance._changed_data:
                    if orig_datum == datum:
                        # Value set back to original value.
                        del instance._changed_data[self.name]
                    else:
                        # Value changed still update to new value.
                        instance._changed_data[self.name] = datum
                elif orig_datum != datum:
                    # New value changed.
                    instance._changed_data[self.name] = datum
            else:
                instance._changed_data[self.name] = datum

    def __delete__(self, instance):
        """Standard Python descriptor method."""
        if self.readonly:
            raise AttributeError("%s is read-only" % self.name)
        elif self.name in instance._data:
            del instance._data[self.name]
            # Mark the field as deleted.
            instance._changed_data[self.name] = None
        else:
            pass  # Nothing to do.


class ObjectFieldRelated(ObjectField):
    def __init__(
        self,
        name,
        cls,
        *,
        default=undefined,
        readonly=False,
        pk=False,
        alt_pk=False,
        reverse=undefined,
        use_data_setter=False,
        map_func=None
    ):
        """Create a `ObjectFieldRelated` with `cls`.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        :param cls: The name of the object class to convert into.
        :param default: A default value to return when `name` is not found in
            the MAAS-side data dictionary.
        :param readonly: If true, prevent setting or deleting of this field.
        :param pk: If true marks the field as the unique primary key for the
            object it is defined on. If an integer then it define its place
            in the tuple of values that makes the object uniquely identified.
        :param alt_pk: If true marks the field as an alternative unique
            primary key for this object it is defined on. If an integer then
            it define its place in the tuple of values that makes the object
            uniquely identified.
        :param reverse: The name of the field on the returned instances of
            `cls` to place this objects instance.
        :param use_data_setter: When True setting the field will use the
            `_data` from the `Object` instead of using the primary keys to
            set on the object.
        """
        super(ObjectFieldRelated, self).__init__(
            name, default=default, readonly=readonly, pk=pk, alt_pk=alt_pk
        )
        self.reverse = reverse
        self.use_data_setter = use_data_setter
        self.map_func = map_func
        if self.map_func is None:
            self.map_func = lambda instance, value: value
        if not isinstance(cls, str):
            if not issubclass(cls, Object):
                raise TypeError("%s is not a subclass of Object" % cls)
            else:
                self.cls = cls.__name__
        else:
            self.cls = cls

    def datum_to_value(self, instance, datum):
        """Convert a given MAAS-side datum to a Python-side value.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The MAAS-side datum to validate and convert into a
            Python-side value.
        :return: A set of `cls` from the given datum.
        """
        datum = self.map_func(instance, datum)
        if datum is None:
            return None
        local_data = None
        if self.reverse is not None:
            local_data = {}
            if self.reverse is undefined:
                local_data[instance.__class__.__name__.lower()] = instance
            else:
                local_data[self.reverse] = instance
        # Get the class from the bound origin.
        bound = getattr(instance._origin, self.cls)
        return bound(datum, local_data=local_data)

    def value_to_datum(self, instance, value):
        """Convert a given Python-side value to a MAAS-side datum.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The Python-side value to validate and convert into a
            MAAS-side datum.
        :return: A datum derived from the given value.
        """
        if value is None:
            return None
        bound = getattr(instance._origin, self.cls)
        if type(value) is bound:
            if self.use_data_setter:
                # Using data setter, so just return the data for the object.
                return value._data
            else:
                # Use the primary keys to set the value.
                descriptors, alt_descriptors = get_pk_descriptors(bound)
                if len(descriptors) == 1:
                    return getattr(value, descriptors[0][0])
                elif len(descriptors) > 1:
                    return tuple(getattr(value, name) for name, _ in descriptors)
                else:
                    raise AttributeError(
                        "unable to perform set object no primary key "
                        "fields defined for %s" % self.cls
                    )
        else:
            raise TypeError("must be %s, not %s" % (self.cls, type(value).__name__))


class ObjectFieldRelatedSet(ObjectField):
    def __init__(
        self, name, cls, *, reverse=undefined, default=undefined, map_func=None
    ):
        """Create a `ObjectFieldRelatedSet` with `cls`.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        :param cls: The name of the object class to convert the sequence of
            data into.
        :param reverse: The name of the field on the returned instances of
            `cls` to place this objects instance.
        :param default: A default value to return when `name` is not found in
            the MAAS-side data dictionary.
        """
        if default is undefined:
            default = []
        super(ObjectFieldRelatedSet, self).__init__(
            name, default=default, readonly=True
        )
        self.reverse = reverse
        self.map_func = map_func
        if self.map_func is None:
            self.map_func = lambda instance, value: value
        if not isinstance(cls, str):
            if not issubclass(cls, ObjectSet):
                raise TypeError("%s is not a subclass of ObjectSet" % cls)
            else:
                self.cls = cls.__name__
        else:
            self.cls = cls

    def datum_to_value(self, instance, datum):
        """Convert a given MAAS-side datum to a Python-side value.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The MAAS-side datum to validate and convert into a
            Python-side value.
        :return: A set of `cls` from the given datum.
        """
        if datum is None:
            return []
        if not isinstance(datum, Sequence):
            raise TypeError("datum must be a sequence, not %s" % type(datum).__name__)
        local_data = None
        if self.reverse is not None:
            local_data = {}
            if self.reverse is undefined:
                local_data[instance.__class__.__name__.lower()] = instance
            else:
                local_data[self.reverse] = instance
        # Get the class from the bound origin.
        bound = getattr(instance._origin, self.cls)
        return bound.Managed(
            instance,
            self,
            (
                bound._object(self.map_func(instance, item), local_data=local_data)
                for item in datum
            ),
        )


class ObjectMethod:
    def __init__(self, _classmethod=None, _instancemethod=None):
        super(ObjectMethod, self).__init__()
        self.__classmethod = _classmethod
        self.__instancemethod = _instancemethod

    def __get__(self, instance, owner):
        if instance is None:
            if self.__classmethod is None:
                raise AttributeError("%s has no matching class attribute" % (instance,))
            else:
                return MethodType(self.__classmethod, owner)
        else:
            if self.__instancemethod is None:
                raise AttributeError(
                    "%s has no matching instance attribute" % (instance,)
                )
            else:
                return MethodType(self.__instancemethod, instance)

    def __set__(self, instance, value):
        # Non-data descriptors (those without __set__) can be shadowed by
        # instance attributes, so prevent that by making this a read-only data
        # descriptor.
        raise AttributeError("%s has no matching instance attribute" % (instance,))

    def classmethod(self, func):
        """Set/modify the class method."""
        self.__classmethod = func
        return self

    @property
    def has_classmethod(self):
        """Has a class method been set?"""
        return self.__classmethod is not None

    def instancemethod(self, func):
        """Set/modify the instance method."""
        self.__instancemethod = func
        return self

    @property
    def has_instancemethod(self):
        """Has an instance method been set?"""
        return self.__instancemethod is not None


class OriginBase:
    """Represents a remote MAAS installation."""

    def __init__(self, session, *, objects=None):
        """
        :param session: A `bones.SessionAPI` instance.
        """
        super(OriginBase, self).__init__()
        self.__session = session
        self.__objects = {} if objects is None else objects
        self.__populate()

    @property
    def session(self):
        """The underlying `bones` session."""
        return self.__session

    def __populate(self):
        # Some objects will not have handlers in the underlying session, but
        # we want to bind them anyway, hence we iterate through all names.
        handlers = dict(self.__session.handlers)
        names = set().union(handlers, self.__objects)
        for name in names:
            handler = handlers.get(name, None)
            base = self.__objects.get(name, Object)
            assert issubclass(type(base), ObjectType)
            obj = base.bind(self, handler, handlers, name=name)
            # Those objects without a custom class are "hidden" by prefixing
            # their name with an underscore.
            objname = "_%s" % name if base is Object else name
            setattr(self, objname, obj)


#
# Conversion/validation functions for use with ObjectTypedField.
#


def to(cls):
    def to_cls(value):
        return cls(value)

    return to_cls


def check(expected):
    def checker(value):
        if issubclass(type(value), expected):
            return value
        else:
            raise TypeError("%r is not of type %s" % (value, expected))

    return checker


def check_optional(expected):
    return check((expected, type(None)))


def parse_timestamp(created):
    created = datetime.strptime(created, "%Y-%m-%dT%H:%M:%S.%f")
    return created.replace(tzinfo=pytz.UTC)


def mapping_of(cls):
    """Expects a mapping from some key to data for `cls` instances."""

    def mapper(data):
        if not isinstance(data, Mapping):
            raise TypeError("data must be a mapping, not %s" % type(data).__name__)
        return {key: cls(value) for key, value in data.items()}

    return mapper


#
# Now it's possible to define the default Origin, which uses a predefined set
# of specialised objects. Most people should use this.
#


def find_objects(modules):
    """Find subclasses of `Object` and `ObjectSet` in the given modules.

    :param modules: The full *names* of modules to include. These modules MUST
        have been imported in advance.
    """
    return {
        subclass.__name__: subclass
        for subclass in chain(get_all_subclasses(Object), get_all_subclasses(ObjectSet))
        if subclass.__module__ in modules
    }


class OriginType(Asynchronous):
    """Metaclass for `Origin`."""

    async def fromURL(cls, url, *, credentials=None, insecure=False):
        """Return an `Origin` for a given MAAS instance."""
        session = await bones.SessionAPI.fromURL(
            url, credentials=credentials, insecure=insecure
        )
        return cls(session)

    def fromProfile(cls, profile):
        """Return an `Origin` from a given configuration profile.

        :see: `ProfileStore`.
        """
        session = bones.SessionAPI.fromProfile(profile)
        return cls(session)

    def fromProfileName(cls, name):
        """Return an `Origin` from a given configuration profile name.

        :see: `ProfileStore`.
        """
        session = bones.SessionAPI.fromProfileName(name)
        return cls(session)

    async def login(cls, url, *, username=None, password=None, insecure=False):
        """Make an `Origin` by logging-in with a username and password.

        :return: A tuple of ``profile`` and ``origin``, where the former is an
            unsaved `Profile` instance, and the latter is an `Origin` instance
            made using the profile.
        """
        profile, session = await bones.SessionAPI.login(
            url=url, username=username, password=password, insecure=insecure
        )
        return profile, cls(session)

    async def connect(cls, url, *, apikey=None, insecure=False):
        """Make an `Origin` by connecting with an apikey.

        :return: A tuple of ``profile`` and ``origin``, where the former is an
            unsaved `Profile` instance, and the latter is an `Origin` instance
            made using the profile.
        """
        profile, session = await bones.SessionAPI.connect(
            url=url, apikey=apikey, insecure=insecure
        )
        return profile, cls(session)

    def __dir__(cls):
        return dir_class(cls)


class Origin(OriginBase, metaclass=OriginType):
    """Represents a remote MAAS installation.

    Uses specialised objects defined in its originating module and specific
    submodules. This is probably the best choice for most people.
    """

    def __init__(self, session):
        """
        :param session: A `bones.SessionAPI` instance.

        """
        modules = {
            ".",
            ".account",
            ".bcache_cache_sets",
            ".bcaches",
            ".block_devices",
            ".boot_resources",
            ".boot_source_selections",
            ".boot_sources",
            ".controllers",
            ".devices",
            ".domains",
            ".events",
            ".subnets",
            ".fabrics",
            ".spaces",
            ".files",
            ".filesystem_groups",
            ".filesystems",
            ".interfaces",
            ".ipranges",
            ".logical_volumes",
            ".maas",
            ".machines",
            ".nodes",
            ".partitions",
            ".pods",
            ".raids",
            ".resource_pools",
            ".spaces",
            ".sshkeys",
            ".static_routes",
            ".subnets",
            ".tags",
            ".users",
            ".version",
            ".vlans",
            ".volume_groups",
            ".zones",
        }
        super(Origin, self).__init__(
            session,
            objects=find_objects(
                {import_module(name, __name__).__name__ for name in modules}
            ),
        )



"""Objects for spaces."""

__all__ = ["Spaces", "Space"]

from . import check, Object, ObjectField, ObjectFieldRelatedSet, ObjectSet, ObjectType


class SpacesType(ObjectType):
    """Metaclass for `Spaces`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(cls, *, name: str = None, description: str = None):
        """
        Create a `Space` in MAAS.

        :param name: The name of the `Space` (optional, will be given a
        default value if not specified).
        :type name: `str`
        :param description: A description of the `Space` (optional).
        :type description: `str`
        :returns: The created Space
        :rtype: `Space`
        """
        params = {}
        if name is not None:
            params["name"] = name
        if description is not None:
            params["description"] = description
        return cls._object(await cls._handler.create(**params))


class Spaces(ObjectSet, metaclass=SpacesType):
    """The set of Spaces stored in MAAS."""


class SpaceType(ObjectType):

    _default_space_id = 0

    async def get_default(cls):
        """Get the 'default' Space for the MAAS."""
        data = await cls._handler.read(id=cls._default_space_id)
        return cls(data)

    async def read(cls, id: int):
        """Get a `Space` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class Space(Object, metaclass=SpaceType):
    """A Space."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    name = ObjectField.Checked("name", check(str), readonly=True, alt_pk=0)

    # description is allowed in the create call and displayed in the UI
    # but never returned by the API.

    vlans = ObjectFieldRelatedSet("vlans", "Vlans", reverse=None)

    async def delete(self):
        """Delete this Space."""
        if self.id == self._origin.Space._default_space_id:
            raise DeleteDefaultSpace("Cannot delete default space.")
        await self._handler.delete(id=self.id)


class DeleteDefaultSpace(Exception):
    """Default space cannot be deleted."""



"""Base object for all filesystem group objects."""

__all__ = ["FilesystemGroup"]

from typing import Sequence

from . import (
    check,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    ObjectSet,
    undefined,
)
from ..enum import BlockDeviceType


def get_device_object(origin, datum):
    device_type = datum.get("type")
    if device_type in [BlockDeviceType.PHYSICAL.value, BlockDeviceType.VIRTUAL.value]:
        return origin.BlockDevice(datum)
    elif device_type == "partition":
        return origin.Partition(datum)
    else:
        raise ValueError("Unknown devices type: %s" % device_type)


class FilesystemGroupDevices(ObjectSet):
    """Devices that make up a `FilesystemGroup`."""


class DevicesField(ObjectFieldRelatedSet):
    """Field for `FilesystemGroupDevices`."""

    def __init__(self, name):
        """Create a `DevicesField`.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        """
        super(ObjectFieldRelatedSet, self).__init__(name, default=[], readonly=True)

    def datum_to_value(self, instance, datum):
        """Convert a given MAAS-side datum to a Python-side value.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The MAAS-side datum to validate and convert into a
            Python-side value.
        :return: A set of `cls` from the given datum.
        """
        if datum is None:
            return []
        if not isinstance(datum, Sequence):
            raise TypeError("datum must be a sequence, not %s" % type(datum).__name__)
        # Get the class from the bound origin.
        bound = getattr(instance._origin, "FilesystemGroupDevices")
        return bound((get_device_object(instance._origin, item) for item in datum))


class DeviceField(ObjectFieldRelated):
    """Field that returns either `BlockDevice` or `Partition`."""

    def __init__(self, name, readonly=False):
        """Create a `DevicesField`.

        :param name: The name of the field. This is the name that's used to
            store the datum in the MAAS-side data dictionary.
        """
        super(ObjectFieldRelated, self).__init__(
            name, default=undefined, readonly=readonly
        )

    def datum_to_value(self, instance, datum):
        """Convert a given MAAS-side datum to a Python-side value.

        :param instance: The `Object` instance on which this field is
            currently operating. This method should treat it as read-only, for
            example to perform validation with regards to other fields.
        :param datum: The MAAS-side datum to validate and convert into a
            Python-side value.
        :return: A set of `cls` from the given datum.
        """
        return get_device_object(instance._origin, datum)


class FilesystemGroup(Object):
    """A filesystem group on a machine.

    Used by `CacheSet`, `Bcache`, `Raid`, and `VolumeGroup`. Never use
    directly.
    """

    node = ObjectFieldRelated("system_id", "Node", readonly=True, pk=0)
    id = ObjectField.Checked("id", check(int), readonly=True, pk=1)
    name = ObjectField.Checked("name", check(str), check(str), alt_pk=1)



"""Objects for version."""

__all__ = ["Version"]

from distutils.version import StrictVersion

from . import Object, ObjectField, ObjectType


def parse_version(version):
    return StrictVersion(version).version


class VersionType(ObjectType):
    """Metaclass for `Version`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(data)


class Version(Object, metaclass=VersionType):
    """MAAS version information."""

    version = ObjectField("version", readonly=True)
    version_info = ObjectField.Checked("version", parse_version, readonly=True)
    subversion = ObjectField("subversion", readonly=True)
    capabilities = ObjectField.Checked("capabilities", frozenset, readonly=True)

    def __repr__(self):
        return "<%s %s %s [%s]>" % (
            self.__class__.__name__,
            self.version,
            self.subversion,
            " ".join(sorted(self.capabilities)),
        )



"""Basic entry points."""

__all__ = ["connect", "login"]

from .utils.maas_async import asynchronous


@asynchronous
async def connect(url, *, apikey=None, insecure=False):
    """Connect to MAAS at `url` using a previously obtained API key.

    :param url: The URL of MAAS, e.g. http://maas.example.com:5240/MAAS/
    :param apikey: The API key to use, e.g.
        SkTvsyHhzkREvvdtNk:Ywn5FvXVupVPvNUhrN:cm3Q2f5naXYPYsrPDPfQy9Q9cUFaEgbM
    :param insecure: Whether to check TLS certificates when using HTTPS.

    :return: A client object.
    """
    from .facade import Client  # Lazy.
    from .viscera import Origin  # Lazy.

    profile, origin = await Origin.connect(url, apikey=apikey, insecure=insecure)
    return Client(origin)


@asynchronous
async def login(url, *, username=None, password=None, insecure=False):
    """Connect to MAAS at `url` with a user name and password.

    :param url: The URL of MAAS, e.g. http://maas.example.com:5240/MAAS/
    :param username: The user name to use, e.g. fred.
    :param password: The user's password.
    :param insecure: Whether to check TLS certificates when using HTTPS.

    :return: A client object.
    """
    from .facade import Client  # Lazy.
    from .viscera import Origin  # Lazy.

    profile, origin = await Origin.login(
        url, username=username, password=password, insecure=insecure
    )
    return Client(origin)



"""Commands for users."""

__all__ = ["register"]

from . import OriginPagedTableCommand, tables


class cmd_users(OriginPagedTableCommand):
    """List users."""

    def execute(self, origin, options, target):
        table = tables.UsersTable()
        return table.render(target, origin.Users.read())


def register(parser):
    """Register profile commands with the given parser."""
    cmd_users.register(parser)



"""Client facade."""

import enum
from functools import update_wrapper


class Facade:
    """Present a simplified API for interacting with MAAS.

    The viscera API separates set-based interactions from those on individual
    objects — e.g. Machines and Machine — which mirrors the way MAAS's API is
    actually constructed, helps to avoid namespace clashes, and makes testing
    cleaner.

    However, we want to present a simplified commingled namespace to users of
    MAAS's *client* API. For example, all entry points related to machines
    should be available as ``client.machines``. This facade class allows us to
    present that commingled namespace without coding it as such.
    """

    def __init__(self, client, name, methods):
        super(Facade, self).__init__()
        self._client = client
        self._name = name
        self._populate(methods)

    def _populate(self, methods):
        for name, func in methods.items():
            setattr(self, name, func)

    def __repr__(self):
        return "<%s>" % self._name


class FacadeDescriptor:
    """Lazily create a facade on first use.

    It will be stored in the instance dictionary using the given name. This
    should match the name by which the descriptor is bound into the instance
    class's namespace: as this is a non-data descriptor [1] this will yield
    create-on-first-use behaviour.

    The factory function should accept a single argument, an `Origin`, and
    return a dict mapping method names to methods of objects obtained from the
    origin.

    [1] https://docs.python.org/3.5/howto/descriptor.html#descriptor-protocol
    """

    def __init__(self, name, factory):
        super(FacadeDescriptor, self).__init__()
        self.name, self.factory = name, factory

    def __get__(self, obj, typ=None):
        methods = self.factory(obj._origin)
        facade = Facade(obj, self.name, methods)
        obj.__dict__[self.name] = facade
        return facade


def facade(factory):
    """Declare a method as a facade factory."""
    wrapper = FacadeDescriptor(factory.__name__, factory)
    return update_wrapper(wrapper, factory)


class Client:
    """A simplified API for interacting with MAAS."""

    def __init__(self, origin):
        super(Client, self).__init__()
        self._origin = origin

    @facade
    def account(origin):
        return {
            "create_credentials": origin.Account.create_credentials,
            "delete_credentials": origin.Account.delete_credentials,
        }

    @facade
    def boot_resources(origin):
        return {
            "create": origin.BootResources.create,
            "get": origin.BootResource.read,
            "list": origin.BootResources.read,
            "start_import": origin.BootResources.start_import,
            "stop_import": origin.BootResources.stop_import,
        }

    @facade
    def boot_sources(origin):
        return {
            "create": origin.BootSources.create,
            "get": origin.BootSource.read,
            "list": origin.BootSources.read,
        }

    @facade
    def devices(origin):
        return {
            "create": origin.Devices.create,
            "get": origin.Device.read,
            "list": origin.Devices.read,
        }

    @facade
    def domains(origin):
        return {
            "create": origin.Domains.create,
            "get": origin.Domain.read,
            "list": origin.Domains.read,
        }

    @facade
    def events(origin):
        namespace = {"query": origin.Events.query}
        namespace.update({level.name: level for level in origin.Events.Level})
        return namespace

    @facade
    def fabrics(origin):
        return {
            "create": origin.Fabrics.create,
            "get": origin.Fabric.read,
            "get_default": origin.Fabric.get_default,
            "list": origin.Fabrics.read,
        }

    @facade
    def pods(origin):
        return {
            "create": origin.Pods.create,
            "list": origin.Pods.read,
            "get": origin.Pod.read,
        }

    @facade
    def static_routes(origin):
        return {
            "create": origin.StaticRoutes.create,
            "get": origin.StaticRoute.read,
            "list": origin.StaticRoutes.read,
        }

    @facade
    def subnets(origin):
        return {
            "create": origin.Subnets.create,
            "get": origin.Subnet.read,
            "list": origin.Subnets.read,
        }

    @facade
    def spaces(origin):
        return {
            "create": origin.Spaces.create,
            "get": origin.Space.read,
            "get_default": origin.Space.get_default,
            "list": origin.Spaces.read,
        }

    @facade
    def files(origin):
        return {"list": origin.Files.read}

    @facade
    def ip_ranges(origin):
        return {
            "create": origin.IPRanges.create,
            "get": origin.IPRange.read,
            "list": origin.IPRanges.read,
        }

    @facade
    def maas(origin):
        attrs = (
            (name, getattr(origin.MAAS, name))
            for name in dir(origin.MAAS)
            if not name.startswith("_")
        )
        return {
            name: attr
            for name, attr in attrs
            if isinstance(attr, enum.EnumMeta) or name.startswith(("get_", "set_"))
        }

    @facade
    def machines(origin):
        return {
            "allocate": origin.Machines.allocate,
            "create": origin.Machines.create,
            "get": origin.Machine.read,
            "list": origin.Machines.read,
            "get_power_parameters_for": origin.Machines.get_power_parameters_for,
        }

    @facade
    def rack_controllers(origin):
        return {"get": origin.RackController.read, "list": origin.RackControllers.read}

    @facade
    def region_controllers(origin):
        return {
            "get": origin.RegionController.read,
            "list": origin.RegionControllers.read,
        }

    @facade
    def ssh_keys(origin):
        return {
            "create": origin.SSHKeys.create,
            "get": origin.SSHKey.read,
            "list": origin.SSHKeys.read,
        }

    @facade
    def tags(origin):
        return {
            "create": origin.Tags.create,
            "get": origin.Tag.read,
            "list": origin.Tags.read,
        }

    @facade
    def users(origin):
        return {
            "create": origin.Users.create,
            "list": origin.Users.read,
            "whoami": origin.Users.whoami,
        }

    @facade
    def version(origin):
        return {"get": origin.Version.read}

    @facade
    def zones(origin):
        return {
            "create": origin.Zones.create,
            "get": origin.Zone.read,
            "list": origin.Zones.read,
        }

    @facade
    def resource_pools(origin):
        return {
            "create": origin.ResourcePools.create,
            "get": origin.ResourcePool.read,
            "list": origin.ResourcePools.read,
        }



"""Objects for boot source selections."""

__all__ = ["BootSourceSelection", "BootSourceSelections"]


from collections.abc import Sequence
from . import check, Object, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType
from .boot_sources import BootSource


class BootSourceSelectionsType(ObjectType):
    """Metaclass for `BootSourceSelections`."""

    async def create(
        cls, boot_source, os, release, *, arches=None, subarches=None, labels=None
    ):
        """Create a new `BootSourceSelection`."""
        if not isinstance(boot_source, BootSource):
            raise TypeError(
                "boot_source must be a BootSource, not %s" % type(boot_source).__name__
            )
        if arches is None:
            arches = ["*"]
        if subarches is None:
            subarches = ["*"]
        if labels is None:
            labels = ["*"]
        data = await cls._handler.create(
            boot_source_id=boot_source.id,
            os=os,
            release=release,
            arches=arches,
            subarches=subarches,
            labels=labels,
        )
        return cls._object(data, {"boot_source_id": boot_source.id})

    async def read(cls, boot_source):
        """Get list of `BootSourceSelection`'s."""
        if isinstance(boot_source, int):
            boot_source_id = boot_source
        elif isinstance(boot_source, BootSource):
            boot_source_id = boot_source.id
        else:
            raise TypeError(
                "boot_source must be a BootSource or int, not %s"
                % type(boot_source).__name__
            )
        data = await cls._handler.read(boot_source_id=boot_source_id)
        return cls(
            cls._object(item, local_data={"boot_source_id": boot_source_id})
            for item in data
        )


class BootSourceSelections(ObjectSet, metaclass=BootSourceSelectionsType):
    """The set of boot source selections."""


class BootSourceSelectionType(ObjectType):
    async def read(cls, boot_source, id):
        """Get `BootSourceSelection` by `id`."""
        if isinstance(boot_source, int):
            boot_source_id = boot_source
        elif isinstance(boot_source, BootSource):
            boot_source_id = boot_source.id
        else:
            raise TypeError(
                "boot_source must be a BootSource or int, not %s"
                % type(boot_source).__name__
            )
        data = await cls._handler.read(boot_source_id=boot_source_id, id=id)
        return cls(data, {"boot_source_id": boot_source_id})


class BootSourceSelection(Object, metaclass=BootSourceSelectionType):
    """A boot source selection."""

    # Only client-side. Classes in this file place `boot_source_id` on
    # the object using `local_data`.
    boot_source = ObjectFieldRelated(
        "boot_source_id", "BootSource", readonly=True, pk=0
    )

    id = ObjectField.Checked("id", check(int), readonly=True, pk=1)
    os = ObjectField.Checked("os", check(str), check(str))
    release = ObjectField.Checked("release", check(str), check(str))
    arches = ObjectField.Checked(  # List[str]
        "arches", check(Sequence), check(Sequence)
    )
    subarches = ObjectField.Checked(  # List[str]
        "subarches", check(Sequence), check(Sequence)
    )
    labels = ObjectField.Checked(  # List[str]
        "labels", check(Sequence), check(Sequence)
    )

    def __repr__(self):
        return super(BootSourceSelection, self).__repr__(
            fields={"os", "release", "arches", "subarches", "labels"}
        )

    async def delete(self):
        """Delete boot source selection."""
        await self._handler.delete(boot_source_id=self.boot_source.id, id=self.id)
